# -*- coding: utf-8 -*-
# @Time    : 2020/12/27 18:09:15 
# @Author  : ddvv
# @Site    : https://ddvvmmzz.github.io
# @File    : ole_infos.py 
# @Software: Visual Studio Code


from data_lib.mmpi.common.abstracts import Signature


class OleTypeNotMatch(Signature):
    authors = ["ddvv"]
    sig_type = 'ole'
    name = "ole_type_not_match"
    severity = 3
    description = "Ole Type Not Match"

    def on_complete(self):
        results = self.get_results()
        for result in results:
            if result.get('type', '') == self.sig_type:
                infos = result.get('value', {}).get('infos', [])
                for info in infos:
                    o_type = info.get('type')
                    m_type = info.get('magic_type')
                    if o_type != m_type:
                        self.mark(type="ole", tag=self.name, o_type=o_type, m_type=m_type)
                return self.has_marks()
        return None


# -*- coding: utf-8 -*-
# @Time    : 2021/01/13 22:56:27 
# @Author  : ddvv
# @Site    : https://ddvvmmzz.github.io
# @File    : lnk.py 
# @Software: Visual Studio Code

# https://github.com/silascutler/LnkParse/blob/master/lnkfile/__init__.py

__description__ = 'Windows Shortcut file (LNK) parser'
__author__ = 'Silas Cutler'
__version__ = '0.2.1'

import json
from logging import raiseExceptions
import struct
import datetime


class lnk_file(object):
	def __init__(self, fhandle=None, indata=None, debug=False):
		self.define_static()

		if fhandle:
			self.indata = fhandle.read()
		elif indata:
			self.indata = indata

		self.debug = debug
		self.lnk_header = {}

		self.linkFlag = {
			'HasTargetIDList': False,
			'HasLinkInfo': False,
			'HasName': False,
			'HasRelativePath': False,
			'HasWorkingDir': False,
			'HasArguments': False,
			'HasIconLocation': False,
			'IsUnicode': False,
			'ForceNoLinkInfo': False,
			'HasExpString': False,
			'RunInSeparateProcess': False,
			'Reserved0': False,
			'HasDarwinID': False,
			'RunAsUser': False,
			'HasExpIcon': False,
			'NoPidlAlias': False,
			'Reserved1': False,
			'RunWithShimLayer': False,
			'ForceNoLinkTrack': False,
			'EnableTargetMetadata': False,
			'DisableLinkPathTracking': False,
			'DisableKnownFolderTracking': False,
			'DisableKnownFolderAlias': False,
			'AllowLinkToLink': False,
			'UnaliasOnSave': False,
			'PreferEnvironmentPath': False,
			'KeepLocalIDListForUNCTarget': False,
		}
		self.fileFlag = {
			'FILE_ATTRIBUTE_READONLY': False,
			'FILE_ATTRIBUTE_HIDDEN': False,
			'FILE_ATTRIBUTE_SYSTEM': False,
			'Reserved, not used by the LNK format': False,
			'FILE_ATTRIBUTE_DIRECTORY': False,
			'FILE_ATTRIBUTE_ARCHIVE': False,
			'FILE_ATTRIBUTE_DEVICE': False,
			'FILE_ATTRIBUTE_NORMAL': False,
			'FILE_ATTRIBUTE_TEMPORARY': False,
			'FILE_ATTRIBUTE_SPARSE_FILE': False,
			'FILE_ATTRIBUTE_REPARSE_POINT': False,
			'FILE_ATTRIBUTE_COMPRESSED': False,
			'FILE_ATTRIBUTE_OFFLINE': False,
			'FILE_ATTRIBUTE_NOT_CONTENT_INDEXED': False,
			'FILE_ATTRIBUTE_ENCRYPTED': False,
			'Unknown (seen on Windows 95 FAT)': False,
			'FILE_ATTRIBUTE_VIRTUAL': False,
		}

		self.targets = {
			'size': 0,
			'items': [],
		}

		self.loc_information = {}
		self.data = {}
		self.extraBlocks = {}
		self.is_lnk = True

		self.process()
		self.define_common()

	def is_lnk_file(self):
		return self.is_lnk

	def define_common(self):
		try:
			out = ''
			if self.linkFlag['HasRelativePath']:
				out += self.data['relativePath']
			if self.linkFlag['HasArguments']:
				out += ' ' + self.data['commandLineArguments']

			self.lnk_command = out
		except Exception as e:
			if self.debug:
				print('Exception define_common: %s' % e)

	def get_command(self):
		try:
			out = ''
			if self.linkFlag['HasRelativePath']:
				out += self.data['relativePath']
			if self.linkFlag['HasArguments']:
				out += ' ' + self.data['commandLineArguments']

			return out
		except Exception as e:
			if self.debug:
				print('Exception get_command: %s' % (e))
			return ''

	def define_static(self):
		# Define static constents used within the LNK format

		# Each MAGIC string refernces a function for processing
		self.EXTRA_SIGS = {
			'a0000001': self.parse_environment_block,
			'a0000002': self.parse_console_block,
			'a0000003': self.parse_distributedTracker_block,
			'a0000004': self.parse_codepage_block,
			'a0000005': self.parse_specialFolder_block,
			'a0000006': self.parse_darwin_block,
			'a0000007': self.parse_icon_block,
			'a0000008': self.parse_shimLayer_block,
			'a0000009': self.parse_metadata_block,
			'a000000b': self.parse_knownFolder_block,
			'a000000c': self.parse_shellItem_block,
		}

		self.DRIVE_TYPES = [
			'DRIVE_UNKNOWN',
			'DRIVE_NO_ROOT_DIR',
			'DRIVE_REMOVABLE',
			'DRIVE_FIXED',
			'DRIVE_REMOTE',
			'DRIVE_CDROM',
			'DRIVE_RAMDISK',
		]
		self.HOTKEY_VALUES = {
			'\x00': 'UNSET',
			'\x01': 'HOTKEYF_SHIFT',
			'\x02': 'HOTKEYF_CONTROL',
			'\x03': 'HOTKEYF_ALT',
		}
		self.WINDOWSTYLES = [
			'SW_HIDE',
			'SW_NORMAL',
			'SW_SHOWMINIMIZED',
			'SW_MAXIMIZE ',
			'SW_SHOWNOACTIVATE',
			'SW_SHOW',
			'SW_MINIMIZE',
			'SW_SHOWMINNOACTIVE',
			'SW_SHOWNA',
			'SW_RESTORE',
			'SW_SHOWDEFAULT',
		]

	@staticmethod
	def clean_line(rstring):
		return ''.join(chr(i) for i in rstring if 128 > i > 20)

	def parse_lnk_header(self):
		# Parse the LNK file header
		try:
			# Header always starts with { 4c 00 00 00 } and is the size of the header
			self.lnk_header['header_size'] = struct.unpack('<I', self.indata[:4])[0]

			lnk_header = self.indata[:self.lnk_header['header_size']]

			self.lnk_header['guid'] = lnk_header[4:20].hex()

			self.lnk_header['rlinkFlags'] = struct.unpack('<i', lnk_header[20:24])[0]
			self.lnk_header['rfileFlags'] = struct.unpack('<i', lnk_header[24:28])[0]

			self.lnk_header['creation_time'] = struct.unpack('<q', lnk_header[28:36])[0]
			self.lnk_header['accessed_time'] = struct.unpack('<q', lnk_header[36:44])[0]
			self.lnk_header['modified_time'] = struct.unpack('<q', lnk_header[44:52])[0]

			self.lnk_header['file_size'] = struct.unpack('<i', lnk_header[52:56])[0]
			self.lnk_header['rfile_size'] = lnk_header[52:56].hex()

			self.lnk_header['icon_index'] = struct.unpack('<I', lnk_header[56:60])[0]
			try:
				if struct.unpack('<i', lnk_header[60:64])[0] < len(self.WINDOWSTYLES):
					self.lnk_header['windowstyle'] = self.WINDOWSTYLES[
						struct.unpack('<i', lnk_header[60:64])[0]]
				else:
					self.lnk_header['windowstyle'] = struct.unpack('<i', lnk_header[60:64])[0]
			except Exception as e:
				if self.debug:
					print('Error Parsing WindowStyle in Header: %s' % e)
				self.lnk_header['windowstyle'] = struct.unpack('<i', lnk_header[60:64])[0]

			try:
				self.lnk_header['hotkey'] = '%s - %s {0x%s}' % (
					self.HOTKEY_VALUES[chr(struct.unpack('<B', lnk_header[65:66])[0])],
					self.clean_line(struct.unpack('<B', lnk_header[64:65])),
					lnk_header[64:66].hex()
				)

				self.lnk_header['rhotkey'] = struct.unpack('<H', lnk_header[64:66])[0]
			except Exception as e:
				if self.debug:
					print('Exception parsing HOTKEY part of header: %s' % e)
					print(lnk_header[65:66].hex())
				self.lnk_header['hotkey'] = struct.unpack('<H', lnk_header[64:66])[0]

			self.lnk_header['reserved0'] = struct.unpack('<H', lnk_header[66:68])[0]
			self.lnk_header['reserved1'] = struct.unpack('<i', lnk_header[68:72])[0]
			self.lnk_header['reserved2'] = struct.unpack('<i', lnk_header[72:76])[0]
		except Exception as e:
			if self.debug:
				print('Exception parsing LNK Header: %s' % e)
			return False

		if self.lnk_header['header_size'] == 76:
			return True

	def parse_link_flags(self):
		if self.lnk_header['rlinkFlags'] & 0x00000001:
			self.linkFlag['HasTargetIDList'] = True
		if self.lnk_header['rlinkFlags'] & 0x00000002:
			self.linkFlag['HasLinkInfo'] = True
		if self.lnk_header['rlinkFlags'] & 0x00000004:
			self.linkFlag['HasName'] = True
		if self.lnk_header['rlinkFlags'] & 0x00000008:
			self.linkFlag['HasRelativePath'] = True
		if self.lnk_header['rlinkFlags'] & 0x00000010:
			self.linkFlag['HasWorkingDir'] = True
		if self.lnk_header['rlinkFlags'] & 0x00000020:
			self.linkFlag['HasArguments'] = True
		if self.lnk_header['rlinkFlags'] & 0x00000040:
			self.linkFlag['HasIconLocation'] = True
		if self.lnk_header['rlinkFlags'] & 0x00000080:
			self.linkFlag['IsUnicode'] = True
		if self.lnk_header['rlinkFlags'] & 0x00000100:
			self.linkFlag['ForceNoLinkInfo'] = True
		if self.lnk_header['rlinkFlags'] & 0x00000200:
			self.linkFlag['HasExpString'] = True
		if self.lnk_header['rlinkFlags'] & 0x00000400:
			self.linkFlag['RunInSeparateProcess'] = True
		if self.lnk_header['rlinkFlags'] & 0x00000800:
			self.linkFlag['Reserved0'] = True
		if self.lnk_header['rlinkFlags'] & 0x00001000:
			self.linkFlag['HasDarwinID'] = True
		if self.lnk_header['rlinkFlags'] & 0x00002000:
			self.linkFlag['RunAsUser'] = True
		if self.lnk_header['rlinkFlags'] & 0x00004000:
			self.linkFlag['HasExpIcon'] = True
		if self.lnk_header['rlinkFlags'] & 0x00008000:
			self.linkFlag['NoPidlAlias'] = True
		if self.lnk_header['rlinkFlags'] & 0x00010000:
			self.linkFlag['Reserved1'] = True
		if self.lnk_header['rlinkFlags'] & 0x00020000:
			self.linkFlag['RunWithShimLayer'] = True
		if self.lnk_header['rlinkFlags'] & 0x00040000:
			self.linkFlag['ForceNoLinkTrack'] = True
		if self.lnk_header['rlinkFlags'] & 0x00080000:
			self.linkFlag['EnableTargetMetadata'] = True
		if self.lnk_header['rlinkFlags'] & 0x00100000:
			self.linkFlag['DisableLinkPathTracking'] = True
		if self.lnk_header['rlinkFlags'] & 0x00200000:
			self.linkFlag['DisableKnownFolderTracking'] = True
		if self.lnk_header['rlinkFlags'] & 0x00400000:
			self.linkFlag['DisableKnownFolderAlias'] = True
		if self.lnk_header['rlinkFlags'] & 0x00800000:
			self.linkFlag['AllowLinkToLink'] = True
		if self.lnk_header['rlinkFlags'] & 0x01000000:
			self.linkFlag['UnaliasOnSave'] = True
		if self.lnk_header['rlinkFlags'] & 0x02000000:
			self.linkFlag['PreferEnvironmentPath'] = True
		if self.lnk_header['rlinkFlags'] & 0x04000000:
			self.linkFlag['KeepLocalIDListForUNCTarget'] = True

		self.lnk_header['linkFlags'] = self.enabled_flags_to_list(self.linkFlag)

	def parse_file_flags(self):
		if self.lnk_header['rfileFlags'] & 0x00000001:
			self.fileFlag['FILE_ATTRIBUTE_READONLY'] = True
		if self.lnk_header['rfileFlags'] & 0x00000002:
			self.fileFlag['FILE_ATTRIBUTE_HIDDEN'] = True
		if self.lnk_header['rfileFlags'] & 0x00000004:
			self.fileFlag['FILE_ATTRIBUTE_SYSTEM'] = True
		if self.lnk_header['rfileFlags'] & 0x00000008:
			self.fileFlag['Reserved, not used by the LNK format'] = True
		if self.lnk_header['rfileFlags'] & 0x00000010:
			self.fileFlag['FILE_ATTRIBUTE_DIRECTORY'] = True
		if self.lnk_header['rfileFlags'] & 0x00000020:
			self.fileFlag['FILE_ATTRIBUTE_ARCHIVE'] = True
		if self.lnk_header['rfileFlags'] & 0x00000040:
			self.fileFlag['FILE_ATTRIBUTE_DEVICE'] = True
		if self.lnk_header['rfileFlags'] & 0x00000080:
			self.fileFlag['FILE_ATTRIBUTE_NORMAL'] = True
		if self.lnk_header['rfileFlags'] & 0x00000100:
			self.fileFlag['FILE_ATTRIBUTE_TEMPORARY'] = True
		if self.lnk_header['rfileFlags'] & 0x00000200:
			self.fileFlag['FILE_ATTRIBUTE_SPARSE_FILE'] = True
		if self.lnk_header['rfileFlags'] & 0x00000400:
			self.fileFlag['FILE_ATTRIBUTE_REPARSE_POINT'] = True
		if self.lnk_header['rfileFlags'] & 0x00000800:
			self.fileFlag['FILE_ATTRIBUTE_COMPRESSED'] = True
		if self.lnk_header['rfileFlags'] & 0x00001000:
			self.fileFlag['FILE_ATTRIBUTE_OFFLINE'] = True
		if self.lnk_header['rfileFlags'] & 0x00002000:
			self.fileFlag['FILE_ATTRIBUTE_NOT_CONTENT_INDEXED'] = True
		if self.lnk_header['rfileFlags'] & 0x00004000:
			self.fileFlag['FILE_ATTRIBUTE_ENCRYPTED'] = True
		if self.lnk_header['rfileFlags'] & 0x00008000:
			self.fileFlag['Unknown (seen on Windows 95 FAT)'] = True
		if self.lnk_header['rfileFlags'] & 0x00010000:
			self.fileFlag['FILE_ATTRIBUTE_VIRTUAL'] = True

		self.lnk_header['fileFlags'] = self.enabled_flags_to_list(self.fileFlag)

	def parse_link_information(self):
		index = 0
		while True:
			tmp_item = {}
			tmp_item['size'] = struct.unpack('<H', self.link_target_list[index: index + 2])[0]
			tmp_item['rsize'] = self.link_target_list[index: index + 2].hex()

			self.items.append(tmp_item)
			index += tmp_item['size']

			return ''

	# Still in development // repair
	def parse_targets(self, index):
		max_size = self.targets['size'] + index

		while (index < max_size):
			ItemID = {
				'size': struct.unpack('<H', self.indata[index: index + 2])[0],
				'type': struct.unpack('<B', self.indata[index + 2: index + 3])[0],
			}
			index += 3

			#           self.targets['items'].append( self.indata[index: index + ItemID['size']].replace('\x00','') )
			#           print('[%s] %s' % (ItemID['size'], hex(ItemID['type']) )#, self.indata[index: index + ItemID['size']].replace('\x00','') ))
			#           print(self.indata[ index: index + ItemID['size'] ].hex()[:50])
			index += ItemID['size']

	#           print(self.indata[index + 2: index + 2 + ItemID['size']].replace('\x00',''))

	def process(self):
		index = 0
		if not self.parse_lnk_header():
			print('Failed Header Check')
			self.is_lnk = False
			return

		self.parse_link_flags()
		self.parse_file_flags()
		index += self.lnk_header['header_size']

		# Parse ID List
		if self.linkFlag['HasTargetIDList']:
			try:
				self.targets['size'] = struct.unpack('<H', self.indata[index: index + 2])[0]
				index += 2
				if self.debug:
					self.parse_targets(index)
				index += self.targets['size']
			except Exception as e:
				if self.debug:
					print('Exception parsing TargetIDList: %s' % e)
				return False

		if self.linkFlag['HasLinkInfo'] and self.linkFlag['ForceNoLinkInfo'] == False:
			try:
				self.loc_information = {
					'LinkInfoSize': struct.unpack('<i', self.indata[index: index + 4])[0],
					'LinkInfoHeaderSize': struct.unpack('<i', self.indata[index + 4: index + 8])[0],
					'LinkInfoFlags': struct.unpack('<i', self.indata[index + 8: index + 12])[0],
					'VolumeIDOffset': struct.unpack('<i', self.indata[index + 12: index + 16])[0],
					'LocalBasePathOffset': struct.unpack('<i', self.indata[index + 16: index + 20])[0],
					'CommonNetworkRelativeLinkOffset': struct.unpack('<i', self.indata[index + 20: index + 24])[0],
					'CommonPathSuffixOffset': struct.unpack('<i', self.indata[index + 24: index + 28])[0],
				}

				if self.loc_information['LinkInfoFlags'] & 0x0001:
					if self.loc_information['LinkInfoHeaderSize'] >= 36:
						self.loc_information['o_LocalBasePathOffsetUnicode'] = \
								struct.unpack('<i', self.indata[index + 28: index + 32])[0]
						local_index = index + self.loc_information['o_LocalBasePathOffsetUnicode']
						self.loc_information['o_LocalBasePathUnicode'] = \
								struct.unpack('<i', self.indata[local_index: local_index + 4])[0]
					else:
						local_index = index + self.loc_information['LocalBasePathOffset']
						self.loc_information['LocalBasePath'] = self.read_string(local_index)

					local_index = index + self.loc_information['VolumeIDOffset']
					self.loc_information['location'] = 'VolumeIDAndLocalBasePath'
					self.loc_information['VolumeIDAndLocalBasePath'] = {
						'VolumeIDSize':
							struct.unpack('<i', self.indata[local_index + 0: local_index + 4])[0],
						'rDriveType':
							struct.unpack('<i', self.indata[local_index + 4: local_index + 8])[0],
						'DriveSerialNumber': hex(
							struct.unpack('<i', self.indata[local_index + 8: local_index + 12])[0]),
						'VolumeLabelOffset':
							struct.unpack('<i', self.indata[local_index + 12: local_index + 16])[0],
					}

					if self.loc_information['VolumeIDAndLocalBasePath']['rDriveType'] < len(self.DRIVE_TYPES):
						self.loc_information['VolumeIDAndLocalBasePath']['DriveType'] = self.DRIVE_TYPES[self.loc_information['VolumeIDAndLocalBasePath']['rDriveType']]

					if self.loc_information['VolumeIDAndLocalBasePath']['VolumeLabelOffset'] != 20:
						length = self.loc_information['VolumeIDAndLocalBasePath']['VolumeIDSize'] - self.loc_information['VolumeIDAndLocalBasePath']['VolumeLabelOffset']
						local_index = index + self.loc_information['VolumeIDOffset'] + self.loc_information['VolumeIDAndLocalBasePath']['VolumeLabelOffset']
						self.loc_information['VolumeIDAndLocalBasePath']['VolumeLabel'] = self.clean_line(self.indata[local_index: local_index + length].replace(b'\x00', b''))
					else:
						self.loc_information['VolumeIDAndLocalBasePath']['o_VolumeLabelOffsetUnicode'] = struct.unpack('<i', self.indata[local_index + 16: local_index + 20])[0]
						local_index = index + self.loc_information['VolumeIDOffset'] + self.loc_information['VolumeIDAndLocalBasePath']['o_VolumeLabelOffsetUnicode']
						self.loc_information['VolumeIDAndLocalBasePath']['o_VolumeLabelUnicode'] = struct.unpack('<i', self.indata[local_index: local_index + 4])[0]

				elif self.loc_information['LinkInfoFlags'] & 0x0002:
					if self.loc_information['LinkInfoHeaderSize'] >= 36:
						self.loc_information['o_CommonPathSuffixOffsetUnicode'] = \
								struct.unpack('<i', self.indata[index + 28: index + 32])[0]
						local_index = index + self.loc_information['o_CommonPathSuffixOffsetUnicode']
						self.loc_information['o_CommonPathSuffixUnicode'] = struct.unpack('<i', self.indata[local_index: local_index + 4])[0]
					else:
						local_index = index + self.loc_information['CommonPathSuffixOffset']
						self.loc_information['CommonPathSuffix'] = \
								struct.unpack('<i', self.indata[local_index: local_index + 4])[0]

					local_index = index + self.loc_information['CommonNetworkRelativeLinkOffset']
					self.loc_information['location'] = 'CommonNetworkRelativeLinkAndPathSuffix'
					self.loc_information['CommonNetworkRelativeLinkAndPathSuffix'] = {
						'CommonNetworkRelativeLinkSize':
							struct.unpack('<i', self.indata[local_index + 0: local_index + 4])[0],
						'CommonNetworkRelativeLinkFlags':
							struct.unpack('<i', self.indata[local_index + 4: local_index + 8])[0],
						'NetNameOffset':
							struct.unpack('<i', self.indata[local_index + 8: local_index + 12])[0],
						'DeviceNameOffset':
							struct.unpack('<i', self.indata[local_index + 12: local_index + 16])[0],
						'NetworkProviderType':
							struct.unpack('<i', self.indata[local_index + 16: local_index + 20])[0],
					}

					if self.loc_information['CommonNetworkRelativeLinkAndPathSuffix']['o_NetNameOffset'] > 20:
						self.loc_information['CommonNetworkRelativeLinkAndPathSuffix']['o_NetNameOffsetUnicode'] = \
						struct.unpack('<i', self.indata[local_index + 20: index + 24])[0]
						local_index = index + self.loc_information['CommonNetworkRelativeLinkAndPathSuffix']['o_NetNameOffsetUnicode']
						self.loc_information['CommonNetworkRelativeLinkAndPathSuffix']['o_NetNameOffsetUnicode'] = \
							struct.unpack('<i', self.indata[local_index: local_index + 4])[0]

						self.loc_information['CommonNetworkRelativeLinkAndPathSuffix']['o_DeviceNameOffsetUnicode'] = \
						struct.unpack('<i', self.indata[local_index + 24: index + 28])[0]
						local_index = self.loc_information['CommonNetworkRelativeLinkAndPathSuffix']['o_DeviceNameOffsetUnicode']
						self.loc_information['CommonNetworkRelativeLinkAndPathSuffix']['o_DeviceNameOffsetUnicode'] = \
							struct.unpack('<i', self.indata[local_index: local_index + 4])[0]
					else:
						local_index = index + self.loc_information['CommonNetworkRelativeLinkAndPathSuffix']['o_NetNameOffset']
						self.loc_information['CommonNetworkRelativeLinkAndPathSuffix']['o_NetNameOffset'] = \
							struct.unpack('<i', self.indata[local_index: local_index + 4])[0]

						local_index = self.loc_information['CommonNetworkRelativeLinkAndPathSuffix']['o_DeviceNameOffset']
						self.loc_information['CommonNetworkRelativeLinkAndPathSuffix']['o_DeviceNameOffset'] = \
							struct.unpack('<i', self.indata[local_index: local_index + 4])[0]

				index += (self.loc_information['LinkInfoSize'])

			except Exception as e:
				if self.debug:
					print('Exception parsing Location information: %s' % e)
				return False

			try:
				u_mult = 1
				if self.linkFlag['IsUnicode']:
					u_mult = 2

				if self.linkFlag['HasName']:
					self.data['description'] = self.read_stringData(index, u_mult)

				if self.linkFlag['HasRelativePath']:
					index, self.data['relativePath'] = self.read_stringData(index, u_mult)

				if self.linkFlag['HasWorkingDir']:
					index, self.data['workingDirectory'] = self.read_stringData(index, u_mult)

				if self.linkFlag['HasArguments']:
					index, self.data['commandLineArguments'] = self.read_stringData(index, u_mult)

				if self.linkFlag['HasIconLocation']:
					index, self.data['iconLocation'] = self.read_stringData(index, u_mult)

			except Exception as e:
				if self.debug:
					print('Exception in parsing data: %s' % e)
				return False

			try:
				while index <= len(self.indata) - 10:
					try:
						size = struct.unpack('<I', self.indata[index: index + 4])[0]
						sig = str(hex(struct.unpack('<I', self.indata[index + 4: index + 8])[0]))[2:]
						self.EXTRA_SIGS[sig](index, size)

						index += (size)
					except Exception as e:
						if self.debug:
							print('Exception in EXTRABLOCK Parsing: %s ' % e)
						index = len(self.data)
						break
			except Exception as e:
				if self.debug:
					print('Exception in EXTRABLOCK: %s' % e)

	def parse_environment_block(self, index, size):
		self.extraBlocks['ENVIRONMENTAL_VARIABLES_LOCATION_BLOCK'] = {}
		self.extraBlocks['ENVIRONMENTAL_VARIABLES_LOCATION_BLOCK']['size'] = size
		self.extraBlocks['ENVIRONMENTAL_VARIABLES_LOCATION_BLOCK'][
			'variable_location'] = self.clean_line(self.indata[index + 8: index + 8 + size])

	def parse_console_block(self, index, size):
		self.extraBlocks['CONSOLE_PROPERTIES_BLOCK'] = {}

	def parse_distributedTracker_block(self, index, size):
		self.extraBlocks['DISTRIBUTED_LINK_TRACKER_BLOCK'] = {}
		self.extraBlocks['DISTRIBUTED_LINK_TRACKER_BLOCK']['size'] = \
			struct.unpack('<I', self.indata[index + 8: index + 12])[0]
		self.extraBlocks['DISTRIBUTED_LINK_TRACKER_BLOCK']['version'] = \
			struct.unpack('<I', self.indata[index + 12: index + 16])[0]

		self.extraBlocks['DISTRIBUTED_LINK_TRACKER_BLOCK']['machine_identifier'] = self.clean_line(
			self.indata[index + 16: index + 32])

		self.extraBlocks['DISTRIBUTED_LINK_TRACKER_BLOCK']['droid_volume_identifier'] = self.indata[
																						index + 32: index + 48].hex()
		self.extraBlocks['DISTRIBUTED_LINK_TRACKER_BLOCK']['droid_file_identifier'] = self.indata[
																					  index + 48: index + 64].hex()
		self.extraBlocks['DISTRIBUTED_LINK_TRACKER_BLOCK'][
			'birth_droid_volume_identifier'] = self.indata[index + 64: index + 80].hex()
		self.extraBlocks['DISTRIBUTED_LINK_TRACKER_BLOCK'][
			'birth_droid_file_identifier'] = self.indata[index + 80: index + 96].hex()

	def parse_codepage_block(self, index, size):
		self.extraBlocks['CONSOLE_CODEPAGE_BLOCK'] = {}

	def parse_specialFolder_block(self, index, size):
		self.extraBlocks['SPECIAL_FOLDER_LOCATION_BLOCK'] = {}

	def parse_darwin_block(self, index, size):
		self.extraBlocks['DARWIN_BLOCK'] = {}

	def parse_icon_block(self, index, size):
		self.extraBlocks['ICON_LOCATION_BLOCK'] = {}

	def parse_shimLayer_block(self, index, size):
		self.extraBlocks['SHIM_LAYER_BLOCK'] = {}

	def parse_metadata_block(self, index, size):
		self.extraBlocks['METADATA_PRPERTIES_BLOCK'] = {}

	def parse_knownFolder_block(self, index, size):
		self.extraBlocks['KNOWN_FOLDER_LOCATION_BLOCK'] = {}

	def parse_shellItem_block(self, index, size):
		self.extraBlocks['SHELL_ITEM_IDENTIFIER_BLOCK'] = {}

	def print_lnk_file(self):
		print('Windows Shortcut Information:')
		print('\tLink Flags: %s - (%s)' % (self.format_linkFlags(), self.lnk_header['rlinkFlags']))
		print('\tFile Flags: %s - (%s)' % (self.format_fileFlags(), self.lnk_header['rfileFlags']))
		print('')
		try:
			print('\tCreation Timestamp: %s' % (self.ms_time_to_unix_time(self.lnk_header['creation_time'])))
			print('\tModified Timestamp: %s' % (self.ms_time_to_unix_time(self.lnk_header['modified_time'])))
			print('\tAccessed Timestamp: %s' % (self.ms_time_to_unix_time(self.lnk_header['accessed_time'])))
			print('')
		except:
			print('\tProblem Parsing Timestamps')
		print(
			'\tFile Size: %s (r: %s)' % (str(self.lnk_header['file_size']), str(len(self.indata))))
		print('\tIcon Index: %s ' % (str(self.lnk_header['icon_index'])))
		print('\tWindow Style: %s ' % (str(self.lnk_header['windowstyle'])))
		print('\tHotKey: %s ' % (str(self.lnk_header['hotkey'])))

		print('')

		for rline in self.data:
			print('\t%s: %s' % (rline, self.data[rline]))

		print('')
		print('\tEXTRA BLOCKS:')
		for enabled in self.extraBlocks:
			print('\t\t%s' % enabled)
			for block in self.extraBlocks[enabled]:
				print('\t\t\t[%s] %s' % (block, self.extraBlocks[enabled][block]))

	def ms_time_to_unix_time(self, time):
		return datetime.datetime.fromtimestamp(time / 10000000.0 - 11644473600).strftime('%Y-%m-%d %H:%M:%S')

	def read_string(self, index):
		result = ''
		while self.indata[index] != 0x00:
			result += chr(self.indata[index])
			index += 1
		return result

	def read_stringData(self, index, u_mult):
		string_size = struct.unpack('<H', self.indata[index: index + 2])[0] * u_mult
		string = self.clean_line(self.indata[index + 2: index + 2 + string_size].replace(b'\x00', b''))
		new_index = index + string_size + 2
		return new_index, string

	@staticmethod
	def enabled_flags_to_list(flags):
		enabled = []
		for flag in flags:
			if flags[flag]:
				enabled.append(flag)
		return enabled

	def format_linkFlags(self):
		enabled = self.enabled_flags_to_list(self.linkFlag)
		return ' | '.join(enabled)

	def format_fileFlags(self):
		enabled = self.enabled_flags_to_list(self.fileFlag)
		return ' | '.join(enabled)

	def print_short(self, pjson=False):
		out = ''
		if self.linkFlag['HasRelativePath']:
			out += self.data['relativePath']
		if self.linkFlag['HasArguments']:
			out += ' ' + self.data['commandLineArguments']

		if pjson:
			print(json.dumps({'command': out}))
		else:
			print(out)

	def get_json(self):
		res = {'header': self.lnk_header, 'data': self.data, 'target': self.targets, 'link_info': self.loc_information, 'extra': self.extraBlocks}

		if 'creation_time' in res['header']:
			res['header']['creation_time'] = self.ms_time_to_unix_time(res['header']['creation_time'])
		if 'accessed_time' in res['header']:
			res['header']['accessed_time'] = self.ms_time_to_unix_time(res['header']['accessed_time'])
		if 'modified_time' in res['header']:
			res['header']['modified_time'] = self.ms_time_to_unix_time(res['header']['modified_time'])

		return res


# -*- coding: utf-8 -*-
# @Time    : 2020/12/20 17:54:58
# @Author  : ddvv
# @Site    : https://ddvvmmzz.github.io
# @File    : filetypes.py
# @Software: Visual Studio Code


FILE_TYPE_MAPPING = {
    'eml': 'eml',
    'html': 'html',
    'zip': 'zip',
    'doc': 'ole',
    'docx': 'zip',
    'xls': 'ole',
    'xlsx': 'zip',
    'ppt': 'ole',
    'pptx': 'zip',
    'rtf': 'rtf',
    'vba': 'vba',
    'exe': 'exe',
    'pdf': 'pdf',
    'ole': 'ole',
    'lnk': 'lnk',
}

OTHER_LIST = ['vba', 'exe', 'pdf', 'bat']

def get_support_file_type(ext):
    return FILE_TYPE_MAPPING.get(ext, 'other')



# -*- coding: utf-8 -*-
# @Time    : 2020/12/20 16:39:47 
# @Author  : ddvv
# @Site    : https://ddvvmmzz.github.io
# @File    : rtf_infos.py 
# @Software: Visual Studio Code


from data_lib.mmpi.common.abstracts import Signature


class RTFSuspiciousDetected(Signature):
    authors = ["ddvv"]
    sig_type = 'rtf'
    name = "rtf_suspicious_detected"
    severity = 3
    description = "RTF Suspicious Detected"

    def on_complete(self):
        results = self.get_results()
        for result in results:
            if result.get('type', '') == self.sig_type:
                infos = result.get('value', {}).get('infos', [])
                for info in infos:
                    if info.get('is_ole', False):
                        self.mark(type="rtf", tag=self.name)
                        return self.has_marks()
        return None


class RTFExploitDetected(Signature):
    authors = ["ddvv"]
    sig_type = 'rtf'
    name = "rtf_exploit_detected"
    severity = 9
    description = "RTF Exploit Detected"

    def on_complete(self):
        results = self.get_results()
        for result in results:
            if result.get('type', '') == self.sig_type:
                infos = result.get('value', {}).get('infos', [])
                for info in infos:
                    if info.get('is_ole', False):
                        class_name = info.get('class_name', '')
                        if class_name == 'OLE2Link' or class_name.lower().startswith('equation'):
                            self.mark(type="rtf", tag=self.name)
                            return self.has_marks()
        return None



# -*- coding: utf-8 -*-
# @Time    : 2020/12/19 11:43:03
# @Author  : ddvv
# @Site    : https://ddvvmmzz.github.io
# @File    : emailparser.py
# @Software: Visual Studio Code


import logging
from datetime import datetime
from hashlib import md5, sha1
from email.message import Message
from email.parser import BytesParser
from email.header import decode_header, Header
from email._parseaddr import AddressList as _AddressList
from email.utils import parsedate_to_datetime

from mmpi.common.abstracts import Processing
from mmpi.common.exceptions import ProcessingError
# from mmpi.common.filetypes import get_magic_type


log = logging.getLogger(__name__)


class EmailParser(Processing):

    @classmethod
    def init_once(cls):
        log.debug("Initializing EmailParser...")

        try:
            cls.ep = BytesParser()
        except Exception as e:
            raise ProcessingError(
                "There was a syntax error in one or more EmailParser: %s" % e
            )

    @classmethod
    def guess_charset(cls, msg):
        charset = msg.get_charset()
        if charset is None:
            content_type = msg.get('Content-Type', '').lower()
            content_type = content_type.replace('"', '')
            pos = content_type.find('charset=')
            if pos >= 0:
                charset = content_type[pos+8:].split(';')[0].strip()
        if charset is None:
            charset = 'utf8'
        if charset == 'cp-850':
            charset = 'cp850'
        return charset

    @classmethod
    def decode_str(cls, s):
        value, charset = decode_header(s)[0]
        if charset:
            if 'utf-8' in charset:
                charset = 'utf-8'
            try:
                value = value.decode(charset, 'ignore')
            except LookupError as e:
                value = value.decode('utf-8', 'ignore')
        return value

    @classmethod
    def calc_md5(cls, data): #计算md5
        data_md5 = md5()
        data_md5.update(data)
        return data_md5.hexdigest()
    
    @classmethod
    def calc_sha1(cls, data): #计算md5
        data_sha1 = sha1()
        data_sha1.update(data)
        return data_sha1.hexdigest()

    def parse_eml_obj(self, emlmsg, info):
        for part in emlmsg.walk():
            if part.get('From'):
                sender = part.get('From')
                sender_list = list()
                recv_list = list()
                subject_text = '', '', '', '', ''
                if isinstance(sender, Header):
                    addrs = _AddressList(sender.__str__()).addresslist
                    for addr in addrs:
                        sender_list.append({'name': self.decode_str(addr[0]), 'addr': addr[1]})
                elif isinstance(sender, str):
                    addrs = _AddressList(sender).addresslist
                    for addr in addrs:
                        sender_list.append({'name': self.decode_str(addr[0]), 'addr': addr[1]})
                to = part.get('To')
                if isinstance(to, Header):
                    addrs = _AddressList(to.__str__()).addresslist
                    for addr in addrs:
                        recv_list.append({'name': self.decode_str(addr[0]), 'addr': addr[1]})
                elif isinstance(to, str):
                    addrs = _AddressList(to).addresslist
                    for addr in addrs:
                        recv_list.append({'name': self.decode_str(addr[0]), 'addr': addr[1]})
                o_date = part.get('Date')
                date = datetime.strptime('1949-10-01 10:08:22', "%Y-%y-%d %H:%M:%S")
                if o_date:
                    try:
                        date = parsedate_to_datetime(o_date)
                    except Exception as e:
                        pass
                X_Originating_IP = part.get('X-Originating-IP')
                if X_Originating_IP:
                    tmp = X_Originating_IP.split(',')
                    tmp[0] = tmp[0][1:]
                    tmp[-1] = tmp[-1][:-1]
                    X_Originating_IP = tmp
                else:
                    X_Originating_IP = []
                subject = part.get('Subject')
                if isinstance(subject, Header):
                    subject_text = self.decode_str(subject.__str__())
                elif isinstance(subject, str):
                    subject_text = self.decode_str(subject)
                header = {'From': sender_list, 'To': recv_list, 'Subject': subject_text, 'Date': date.strftime('%Y-%m-%d %H:%M:%S %Z'), 'X-Originating-IP': X_Originating_IP}
                info['infos'].append(header)
                info['report']['headers'].append(header)
            content_type = part.get('Content-Type')
            if content_type:
                if isinstance(content_type, Header):
                    content_type = content_type.__str__()
                content_type = content_type.lower()
                if content_type.startswith('text/plain'):
                    try:
                        charset = self.guess_charset(part)
                        content = part.get_payload(
                            decode=True).decode(charset, "ignore")
                    except Exception as e:
                        content = part.get_payload()
                    info['datas'].append(
                        {'file_type': 'text', 'content': content})
                    info['report']['body'].append(
                        {'type': 'text', 'content': content})
                elif content_type.startswith('text/html'):
                    try:
                        charset = self.guess_charset(part)
                        content = part.get_payload(
                            decode=True).decode(charset, "ignore")
                    except Exception as e:
                        content = part.get_payload()
                    info['datas'].append(
                        {'file_type': 'html', 'content': content})
            cdispo = part.get('Content-Disposition')
            if cdispo:
                if isinstance(cdispo, Header):
                    cdispo = cdispo.__str__()
                cdispo = cdispo.lower()
                if cdispo.startswith('attachment'):
                    try:
                        # Here is the logic table for this code, based on the email5.0.0 code:
                        #   i     decode  is_multipart  result
                        # ------  ------  ------------  ------------------------------
                        #  None   True    True          None    *
                        #   i     True    True          None
                        #  None   False   True          _payload (a list)
                        #   i     False   True          _payload element i (a Message)
                        #   i     False   False         error (not a list)
                        #   i     True    False         error (not a list)
                        #  None   False   False         _payload
                        #  None   True    False         _payload decoded (bytes) *
                        # in mmpi, just two way use it, marked as *
                        content = part.get_payload(decode=True)
                        if content:
                            attachment_name = part.get_filename()
                            if attachment_name:
                                name = self.decode_str(attachment_name)
                                file_type = name.split('.')[-1]
                                info['datas'].append({'file_type': file_type, 'filename': name, 'filesize': len(content), 'content': content})
                                file_md5 = self.calc_md5(content)
                                file_sha1 = self.calc_sha1(content)
                                # magic_tyep = get_magic_type(content)
                                info['report']['attachments'].append({'type': file_type, 'filename': name, 'filesize': len(content), 'md5': file_md5, 'sha1': file_sha1})
                            else:
                                try:
                                    tmp_msg = self.ep.parsebytes(content)
                                    self.parse_eml_obj(tmp_msg, info)
                                except Exception as e:
                                    log.error(e)
                        else:
                            # if type is eml, something to do
                            contents = part.get_payload()
                            if contents:
                                for content in contents:
                                    if isinstance(content, Message):
                                        self.parse_eml_obj(content, info)
                    except Exception as e:
                        log.error(e)
                        continue


    def run(self):
        self.key = "eml"
        if self.key == self.file_type:
            report = {"headers": list(), "body": list(), "attachments": list()}
            info = {'infos': list(), 'datas': list(), 'report': report}

            if self.file_content:
                emlmsg = self.ep.parsebytes(self.file_content)
                self.parse_eml_obj(emlmsg, info)
            if info.get('infos', []) or info.get('datas', []):
                return info
        return None



# -*- coding: utf-8 -*-
# @Time    : 2020/12/24 23:34:22 
# @Author  : ddvv
# @Site    : https://ddvvmmzz.github.io
# @File    : __init__.py 
# @Software: Visual Studio Code


import os
from urllib.parse import urlparse


def read_domains(filename):
    with open(filename, 'r') as target:
        data = target.readlines()
        return [d.strip() for d in data]

basedir = os.path.dirname(__file__)
filter_file = os.path.join(basedir, 'filter_domains.txt')
short_file = os.path.join(basedir, 'short_domains.txt')

FILTER_DOMAINS = read_domains(filter_file)
SHORT_DOMAINS = read_domains(short_file)


def check_fitler_domain(http_url):
    """
    domain过滤检测，仅检测二级或者三级域名
    返回值：
    True：过滤
    False：不过滤单
    """
    uri = urlparse(http_url)
    domain_name = f"{uri.netloc}"
    domain_list = domain_name.split(':')[0].split('.')
    if len(domain_list) < 2:
        return True
    elif len(domain_list) > 2 and domain_list[-2] in ['com', 'org', 'net', 'int', 'edu', 'gov', 'co']:
        # 取三级域名
        new_domain_name = '.'.join(domain_list[-3:])
    else:
        # 取二级域名
        new_domain_name = '.'.join(domain_list[-2:])
    if new_domain_name in FILTER_DOMAINS:
            return True
    return False


# -*- coding: utf-8 -*-
# @Time    : 2020/12/21 23:02:27
# @Author  : ddvv
# @Site    : https://ddvvmmzz.github.io
# @File    : olevba.py
# @Software: Visual Studio Code


import struct
import math
import logging
import olefile
from io import BytesIO

from mmpi.common import codepages

MODULE_EXTENSION = "bas"
CLASS_EXTENSION = "cls"
FORM_EXTENSION = "frm"

# === LOGGING =================================================================


def get_logger(name, level=logging.CRITICAL+1):
    """
    Create a suitable logger object for this module.
    The goal is not to change settings of the root logger, to avoid getting
    other modules' logs on the screen.
    If a logger exists with same name, reuse it. (Else it would have duplicate
    handlers and messages would be doubled.)
    The level is set to CRITICAL+1 by default, to avoid any logging.
    """
    # First, test if there is already a logger with the same name, else it
    # will generate duplicate messages (due to duplicate handlers):
    if name in logging.Logger.manager.loggerDict:
        # NOTE: another less intrusive but more "hackish" solution would be to
        # use getLogger then test if its effective level is not default.
        logger = logging.getLogger(name)
        # make sure level is OK:
        logger.setLevel(level)
        return logger
    # get a new logger:
    logger = logging.getLogger(name)
    # only add a NullHandler for this logger, it is up to the application
    # to configure its own logging:
    logger.addHandler(logging.NullHandler())
    logger.setLevel(level)
    return logger


# a global logger object used for debugging:
log = get_logger('olevba')


class OlevbaBaseException(Exception):
    """ Base class for exceptions produced here for simpler except clauses """

    def __init__(self, msg, filename=None, orig_exc=None, **kwargs):
        if orig_exc:
            super(OlevbaBaseException, self).__init__(msg +
                                                      ' ({0})'.format(
                                                          orig_exc),
                                                      **kwargs)
        else:
            super(OlevbaBaseException, self).__init__(msg, **kwargs)
        self.msg = msg
        self.filename = filename
        self.orig_exc = orig_exc


class FileOpenError(OlevbaBaseException):
    """ raised by VBA_Parser constructor if all open_... attempts failed

    probably means the file type is not supported
    """

    def __init__(self, filename, orig_exc=None):
        super(FileOpenError, self).__init__(
            'Failed to open file %s' % filename, filename, orig_exc)


class ProcessingError(OlevbaBaseException):
    """ raised by VBA_Parser.process_file* functions """

    def __init__(self, filename, orig_exc):
        super(ProcessingError, self).__init__(
            'Error processing file %s' % filename, filename, orig_exc)


class MsoExtractionError(RuntimeError, OlevbaBaseException):
    """ raised by mso_file_extract if parsing MSO/ActiveMIME data failed """

    def __init__(self, msg):
        MsoExtractionError.__init__(self, msg)
        OlevbaBaseException.__init__(self, msg)


class SubstreamOpenError(FileOpenError):
    """ special kind of FileOpenError: file is a substream of original file """

    def __init__(self, filename, subfilename, orig_exc=None):
        super(SubstreamOpenError, self).__init__(
            str(filename) + '/' + str(subfilename), orig_exc)
        self.filename = filename   # overwrite setting in OlevbaBaseException
        self.subfilename = subfilename


class UnexpectedDataError(OlevbaBaseException):
    """ raised when parsing is strict (=not relaxed) and data is unexpected """

    def __init__(self, stream_path, variable, expected, value):
        if isinstance(expected, int):
            es = '{0:04X}'.format(expected)
        elif isinstance(expected, tuple):
            es = ','.join('{0:04X}'.format(e) for e in expected)
            es = '({0})'.format(es)
        else:
            raise ValueError(
                'Unknown type encountered: {0}'.format(type(expected)))
        super(UnexpectedDataError, self).__init__(
            'Unexpected value in {0} for variable {1}: '
            'expected {2} but found {3:04X}!'
            .format(stream_path, variable, es, value))
        self.stream_path = stream_path
        self.variable = variable
        self.expected = expected
        self.value = value


def copytoken_help(decompressed_current, decompressed_chunk_start):
    """
    compute bit masks to decode a CopyToken according to MS-OVBA 2.4.1.3.19.1 CopyToken Help

    decompressed_current: number of decompressed bytes so far, i.e. len(decompressed_container)
    decompressed_chunk_start: offset of the current chunk in the decompressed container
    return length_mask, offset_mask, bit_count, maximum_length
    """
    difference = decompressed_current - decompressed_chunk_start
    bit_count = int(math.ceil(math.log(difference, 2)))
    bit_count = max([bit_count, 4])
    length_mask = 0xFFFF >> bit_count
    offset_mask = ~length_mask
    maximum_length = (0xFFFF >> bit_count) + 3
    return length_mask, offset_mask, bit_count, maximum_length


def decompress_stream(compressed_container):
    """
    Decompress a stream according to MS-OVBA section 2.4.1

    :param compressed_container bytearray: bytearray or bytes compressed according to the MS-OVBA 2.4.1.3.6 Compression algorithm
    :return: the decompressed container as a bytes string
    :rtype: bytes
    """
    # 2.4.1.2 State Variables

    # The following state is maintained for the CompressedContainer (section 2.4.1.1.1):
    # CompressedRecordEnd: The location of the byte after the last byte in the CompressedContainer (section 2.4.1.1.1).
    # CompressedCurrent: The location of the next byte in the CompressedContainer (section 2.4.1.1.1) to be read by
    #                    decompression or to be written by compression.

    # The following state is maintained for the current CompressedChunk (section 2.4.1.1.4):
    # CompressedChunkStart: The location of the first byte of the CompressedChunk (section 2.4.1.1.4) within the
    #                       CompressedContainer (section 2.4.1.1.1).

    # The following state is maintained for a DecompressedBuffer (section 2.4.1.1.2):
    # DecompressedCurrent: The location of the next byte in the DecompressedBuffer (section 2.4.1.1.2) to be written by
    #                      decompression or to be read by compression.
    # DecompressedBufferEnd: The location of the byte after the last byte in the DecompressedBuffer (section 2.4.1.1.2).

    # The following state is maintained for the current DecompressedChunk (section 2.4.1.1.3):
    # DecompressedChunkStart: The location of the first byte of the DecompressedChunk (section 2.4.1.1.3) within the
    #                         DecompressedBuffer (section 2.4.1.1.2).

    # Check the input is a bytearray, otherwise convert it (assuming it's bytes):
    if not isinstance(compressed_container, bytearray):
        compressed_container = bytearray(compressed_container)
        # raise TypeError('decompress_stream requires a bytearray as input')
    decompressed_container = bytearray()  # result
    compressed_current = 0

    sig_byte = compressed_container[compressed_current]
    if sig_byte != 0x01:
        raise ValueError('invalid signature byte {0:02X}'.format(sig_byte))

    compressed_current += 1

    # NOTE: the definition of CompressedRecordEnd is ambiguous. Here we assume that
    # CompressedRecordEnd = len(compressed_container)
    while compressed_current < len(compressed_container):
        # 2.4.1.1.5
        compressed_chunk_start = compressed_current
        # chunk header = first 16 bits
        compressed_chunk_header = \
            struct.unpack(
                "<H", compressed_container[compressed_chunk_start:compressed_chunk_start + 2])[0]
        # chunk size = 12 first bits of header + 3
        chunk_size = (compressed_chunk_header & 0x0FFF) + 3
        # chunk signature = 3 next bits - should always be 0b011
        chunk_signature = (compressed_chunk_header >> 12) & 0x07
        if chunk_signature != 0b011:
            raise ValueError(
                'Invalid CompressedChunkSignature in VBA compressed stream')
        # chunk flag = next bit - 1 == compressed, 0 == uncompressed
        chunk_flag = (compressed_chunk_header >> 15) & 0x01
        log.debug("chunk size = {0}, compressed flag = {1}".format(
            chunk_size, chunk_flag))

        # MS-OVBA 2.4.1.3.12: the maximum size of a chunk including its header is 4098 bytes (header 2 + data 4096)
        # The minimum size is 3 bytes
        # NOTE: there seems to be a typo in MS-OVBA, the check should be with 4098, not 4095 (which is the max value
        # in chunk header before adding 3.
        # Also the first test is not useful since a 12 bits value cannot be larger than 4095.
        if chunk_flag == 1 and chunk_size > 4098:
            raise ValueError(
                'CompressedChunkSize > 4098 but CompressedChunkFlag == 1')
        if chunk_flag == 0 and chunk_size != 4098:
            raise ValueError(
                'CompressedChunkSize != 4098 but CompressedChunkFlag == 0')

        # check if chunk_size goes beyond the compressed data, instead of silently cutting it:
        # TODO: raise an exception?
        if compressed_chunk_start + chunk_size > len(compressed_container):
            log.warning('Chunk size is larger than remaining compressed data')
        compressed_end = min(
            [len(compressed_container), compressed_chunk_start + chunk_size])
        # read after chunk header:
        compressed_current = compressed_chunk_start + 2

        if chunk_flag == 0:
            # MS-OVBA 2.4.1.3.3 Decompressing a RawChunk
            # uncompressed chunk: read the next 4096 bytes as-is
            # TODO: check if there are at least 4096 bytes left
            decompressed_container.extend(
                compressed_container[compressed_current:compressed_current + 4096])
            compressed_current += 4096
        else:
            # MS-OVBA 2.4.1.3.2 Decompressing a CompressedChunk
            # compressed chunk
            decompressed_chunk_start = len(decompressed_container)
            while compressed_current < compressed_end:
                # MS-OVBA 2.4.1.3.4 Decompressing a TokenSequence
                # log.debug('compressed_current = %d / compressed_end = %d' % (compressed_current, compressed_end))
                # FlagByte: 8 bits indicating if the following 8 tokens are either literal (1 byte of plain text) or
                # copy tokens (reference to a previous literal token)
                flag_byte = compressed_container[compressed_current]
                compressed_current += 1
                for bit_index in range(0, 8):
                    # log.debug('bit_index=%d / compressed_current=%d / compressed_end=%d' % (bit_index, compressed_current, compressed_end))
                    if compressed_current >= compressed_end:
                        break
                    # MS-OVBA 2.4.1.3.5 Decompressing a Token
                    # MS-OVBA 2.4.1.3.17 Extract FlagBit
                    flag_bit = (flag_byte >> bit_index) & 1
                    # log.debug('bit_index=%d: flag_bit=%d' % (bit_index, flag_bit))
                    if flag_bit == 0:  # LiteralToken
                        # copy one byte directly to output
                        decompressed_container.extend(
                            [compressed_container[compressed_current]])
                        compressed_current += 1
                    else:  # CopyToken
                        # MS-OVBA 2.4.1.3.19.2 Unpack CopyToken
                        copy_token = \
                            struct.unpack(
                                "<H", compressed_container[compressed_current:compressed_current + 2])[0]
                        # TODO: check this
                        length_mask, offset_mask, bit_count, _ = copytoken_help(
                            len(decompressed_container), decompressed_chunk_start)
                        length = (copy_token & length_mask) + 3
                        temp1 = copy_token & offset_mask
                        temp2 = 16 - bit_count
                        offset = (temp1 >> temp2) + 1
                        # log.debug('offset=%d length=%d' % (offset, length))
                        copy_source = len(decompressed_container) - offset
                        for index in range(copy_source, copy_source + length):
                            decompressed_container.extend(
                                [decompressed_container[index]])
                        compressed_current += 2
    return bytes(decompressed_container)


class VBA_Module(object):
    """
    Class to parse a VBA module from an OLE file, and to store all the corresponding
    metadata and VBA source code.
    """

    def __init__(self, project, dir_stream, module_index):
        """
        Parse a VBA Module record from the dir stream of a VBA project.
        Reference: MS-OVBA 2.3.4.2.3.2 MODULE Record

        :param VBA_Project project: VBA_Project, corresponding VBA project
        :param olefile.OleStream dir_stream: olefile.OleStream, file object containing the module record
        :param int module_index: int, index of the module in the VBA project list
        """
        #: reference to the VBA project for later use (VBA_Project)
        self.project = project
        #: VBA module name (unicode str)
        self.name = None
        #: VBA module name as a native str (utf8 bytes on py2, str on py3)
        self.name_str = None
        #: VBA module name, unicode copy (unicode str)
        self._name_unicode = None
        #: Stream name containing the VBA module (unicode str)
        self.streamname = None
        #: Stream name containing the VBA module as a native str (utf8 bytes on py2, str on py3)
        self.streamname_str = None
        self._streamname_unicode = None
        self.docstring = None
        self._docstring_unicode = None
        self.textoffset = None
        self.type = None
        self.readonly = False
        self.private = False
        #: VBA source code in bytes format, using the original code page from the VBA project
        self.code_raw = None
        #: VBA source code in unicode format (unicode for Python2, str for Python 3)
        self.code = None
        #: VBA source code in native str format (str encoded with UTF-8 for Python 2, str for Python 3)
        self.code_str = None
        #: VBA module file name including an extension based on the module type such as bas, cls, frm (unicode str)
        self.filename = None
        #: VBA module file name in native str format (str)
        self.filename_str = None
        self.code_path = None
        try:
            # 2.3.4.2.3.2.1 MODULENAME Record
            # Specifies a VBA identifier as the name of the containing MODULE Record
            _id = struct.unpack("<H", dir_stream.read(2))[0]
            project.check_value('MODULENAME_Id', 0x0019, _id)
            size = struct.unpack("<L", dir_stream.read(4))[0]
            modulename_bytes = dir_stream.read(size)
            # Module name always stored as Unicode:
            self.name = project.decode_bytes(modulename_bytes)
            self.name_str = self.name
            # account for optional sections
            # TODO: shouldn't this be a loop? (check MS-OVBA)
            section_id = struct.unpack("<H", dir_stream.read(2))[0]
            if section_id == 0x0047:
                # 2.3.4.2.3.2.2 MODULENAMEUNICODE Record
                # Specifies a VBA identifier as the name of the containing MODULE Record (section 2.3.4.2.3.2).
                # MUST contain the UTF-16 encoding of MODULENAME Record
                size = struct.unpack("<L", dir_stream.read(4))[0]
                self._name_unicode = dir_stream.read(
                    size).decode('UTF-16LE', 'replace')
                section_id = struct.unpack("<H", dir_stream.read(2))[0]
            if section_id == 0x001A:
                # 2.3.4.2.3.2.3 MODULESTREAMNAME Record
                # Specifies the stream name of the ModuleStream (section 2.3.4.3) in the VBA Storage (section 2.3.4)
                # corresponding to the containing MODULE Record
                size = struct.unpack("<L", dir_stream.read(4))[0]
                streamname_bytes = dir_stream.read(size)
                # Store it as Unicode:
                self.streamname = project.decode_bytes(streamname_bytes)
                self.streamname_str = self.streamname
                reserved = struct.unpack("<H", dir_stream.read(2))[0]
                project.check_value(
                    'MODULESTREAMNAME_Reserved', 0x0032, reserved)
                size = struct.unpack("<L", dir_stream.read(4))[0]
                self._streamname_unicode = dir_stream.read(
                    size).decode('UTF-16LE', 'replace')
                section_id = struct.unpack("<H", dir_stream.read(2))[0]
            if section_id == 0x001C:
                # 2.3.4.2.3.2.4 MODULEDOCSTRING Record
                # Specifies the description for the containing MODULE Record
                size = struct.unpack("<L", dir_stream.read(4))[0]
                docstring_bytes = dir_stream.read(size)
                self.docstring = project.decode_bytes(docstring_bytes)
                reserved = struct.unpack("<H", dir_stream.read(2))[0]
                project.check_value(
                    'MODULEDOCSTRING_Reserved', 0x0048, reserved)
                size = struct.unpack("<L", dir_stream.read(4))[0]
                self._docstring_unicode = dir_stream.read(size)
                section_id = struct.unpack("<H", dir_stream.read(2))[0]
            if section_id == 0x0031:
                # 2.3.4.2.3.2.5 MODULEOFFSET Record
                # Specifies the location of the source code within the ModuleStream (section 2.3.4.3)
                # that corresponds to the containing MODULE Record
                size = struct.unpack("<L", dir_stream.read(4))[0]
                project.check_value('MODULEOFFSET_Size', 0x0004, size)
                self.textoffset = struct.unpack("<L", dir_stream.read(4))[0]
                section_id = struct.unpack("<H", dir_stream.read(2))[0]
            if section_id == 0x001E:
                # 2.3.4.2.3.2.6 MODULEHELPCONTEXT Record
                # Specifies the Help topic identifier for the containing MODULE Record
                modulehelpcontext_size = struct.unpack(
                    "<L", dir_stream.read(4))[0]
                project.check_value('MODULEHELPCONTEXT_Size',
                                    0x0004, modulehelpcontext_size)
                # HelpContext (4 bytes): An unsigned integer that specifies the Help topic identifier
                # in the Help file specified by PROJECTHELPFILEPATH Record
                helpcontext = struct.unpack("<L", dir_stream.read(4))[0]
                section_id = struct.unpack("<H", dir_stream.read(2))[0]
            if section_id == 0x002C:
                # 2.3.4.2.3.2.7 MODULECOOKIE Record
                # Specifies ignored data.
                size = struct.unpack("<L", dir_stream.read(4))[0]
                project.check_value('MODULECOOKIE_Size', 0x0002, size)
                cookie = struct.unpack("<H", dir_stream.read(2))[0]
                section_id = struct.unpack("<H", dir_stream.read(2))[0]
            if section_id == 0x0021 or section_id == 0x0022:
                # 2.3.4.2.3.2.8 MODULETYPE Record
                # Specifies whether the containing MODULE Record (section 2.3.4.2.3.2) is a procedural module,
                # document module, class module, or designer module.
                # Id (2 bytes): An unsigned integer that specifies the identifier for this record.
                # MUST be 0x0021 when the containing MODULE Record (section 2.3.4.2.3.2) is a procedural module.
                # MUST be 0x0022 when the containing MODULE Record (section 2.3.4.2.3.2) is a document module,
                # class module, or designer module.
                self.type = section_id
                reserved = struct.unpack("<L", dir_stream.read(4))[0]
                section_id = struct.unpack("<H", dir_stream.read(2))[0]
            if section_id == 0x0025:
                # 2.3.4.2.3.2.9 MODULEREADONLY Record
                # Specifies that the containing MODULE Record (section 2.3.4.2.3.2) is read-only.
                self.readonly = True
                reserved = struct.unpack("<L", dir_stream.read(4))[0]
                project.check_value(
                    'MODULEREADONLY_Reserved', 0x0000, reserved)
                section_id = struct.unpack("<H", dir_stream.read(2))[0]
            if section_id == 0x0028:
                # 2.3.4.2.3.2.10 MODULEPRIVATE Record
                # Specifies that the containing MODULE Record (section 2.3.4.2.3.2) is only usable from within
                # the current VBA project.
                self.private = True
                reserved = struct.unpack("<L", dir_stream.read(4))[0]
                project.check_value('MODULEPRIVATE_Reserved', 0x0000, reserved)
                section_id = struct.unpack("<H", dir_stream.read(2))[0]
            if section_id == 0x002B:  # TERMINATOR
                # Terminator (2 bytes): An unsigned integer that specifies the end of this record. MUST be 0x002B.
                # Reserved (4 bytes): MUST be 0x00000000. MUST be ignored.
                reserved = struct.unpack("<L", dir_stream.read(4))[0]
                project.check_value('MODULE_Reserved', 0x0000, reserved)
                section_id = None
            if section_id != None:
                log.warning(
                    'unknown or invalid module section id {0:04X}'.format(section_id))

            log.debug("Module Name = {0}".format(self.name_str))
            # log.debug("Module Name Unicode = {0}".format(self._name_unicode))
            log.debug("Stream Name = {0}".format(self.streamname_str))
            # log.debug("Stream Name Unicode = {0}".format(self._streamname_unicode))
            log.debug("TextOffset = {0}".format(self.textoffset))

            code_data = None
            # let's try the different names we have, just in case some are missing:
            try_names = (self.streamname, self._streamname_unicode,
                         self.name, self._name_unicode)
            for stream_name in try_names:
                # TODO: if olefile._find were less private, could replace this
                #        try-except with calls to it
                if stream_name is not None:
                    try:
                        self.code_path = project.vba_root + u'VBA/' + stream_name
                        log.debug('opening VBA code stream %s' %
                                  self.code_path)
                        code_data = project.ole.openstream(
                            self.code_path).read()
                        break
                    except IOError as ioe:
                        log.debug('failed to open stream VBA/%r (%r), try other name'
                                  % (stream_name, ioe))

            if code_data is None:
                log.info("Could not open stream %d of %d ('VBA/' + one of %r)!"
                         % (module_index, project.modules_count,
                            '/'.join("'" + stream_name + "'"
                                     for stream_name in try_names)))
                if project.relaxed:
                    return  # ... continue with next submodule
                else:
                    raise SubstreamOpenError('[BASE]', 'VBA/' + self.name)

            log.debug("length of code_data = {0}".format(len(code_data)))
            log.debug("offset of code_data = {0}".format(self.textoffset))
            code_data = code_data[self.textoffset:]
            if len(code_data) > 0:
                code_data = decompress_stream(bytearray(code_data))
                # store the raw code encoded as bytes with the project's code page:
                self.code_raw = code_data
                # decode it to unicode:
                self.code = project.decode_bytes(code_data)
                # also store a native str version:
                self.code_str = self.code
                # case-insensitive search in the code_modules dict to find the file extension:
                filext = self.project.module_ext.get(self.name.lower(), 'vba')
                self.filename = u'{0}.{1}'.format(self.name, filext)
                self.filename_str = self.filename
                log.debug('extracted file {0}'.format(self.filename_str))
            else:
                log.warning("module stream {0} has code data length 0".format(
                    self.streamname_str))
        except (UnexpectedDataError, SubstreamOpenError):
            raise
        except Exception as exc:
            log.info('Error parsing module {0} of {1}:'
                     .format(module_index, project.modules_count),
                     exc_info=True)
            if not project.relaxed:
                raise


class VBA_Project(object):
    """
    Class to parse a VBA project from an OLE file, and to store all the corresponding
    metadata and VBA modules.
    """

    def __init__(self, ole, vba_root, project_path, dir_path, relaxed=True):
        """
        Extract VBA macros from an OleFileIO object.

        :param vba_root: path to the VBA root storage, containing the VBA storage and the PROJECT stream
        :param project_path: path to the PROJECT stream
        :param relaxed: If True, only create info/debug log entry if data is not as expected
                        (e.g. opening substream fails); if False, raise an error in this case
        """
        self.ole = ole
        self.vba_root = vba_root
        self. project_path = project_path
        self.dir_path = dir_path
        self.relaxed = relaxed
        #: VBA modules contained in the project (list of VBA_Module objects)
        self.modules = []
        #: file extension for each VBA module
        self.module_ext = {}
        log.debug('Parsing the dir stream from %r' % dir_path)
        # read data from dir stream (compressed)
        dir_compressed = ole.openstream(dir_path).read()
        # decompress it:
        dir_stream = BytesIO(decompress_stream(bytearray(dir_compressed)))
        # store reference for later use:
        self.dir_stream = dir_stream

        # reference: MS-VBAL 2.3.4.2 dir Stream: Version Independent Project Information

        # PROJECTSYSKIND Record
        # Specifies the platform for which the VBA project is created.
        projectsyskind_id = struct.unpack("<H", dir_stream.read(2))[0]
        self.check_value('PROJECTSYSKIND_Id', 0x0001, projectsyskind_id)
        projectsyskind_size = struct.unpack("<L", dir_stream.read(4))[0]
        self.check_value('PROJECTSYSKIND_Size', 0x0004, projectsyskind_size)
        self.syskind = struct.unpack("<L", dir_stream.read(4))[0]
        SYSKIND_NAME = {
            0x00: "16-bit Windows",
            0x01: "32-bit Windows",
            0x02: "Macintosh",
            0x03: "64-bit Windows"
        }
        self.syskind_name = SYSKIND_NAME.get(self.syskind, 'Unknown')
        log.debug("PROJECTSYSKIND_SysKind: %d - %s" %
                  (self.syskind, self.syskind_name))
        if self.syskind not in SYSKIND_NAME:
            log.error(
                "invalid PROJECTSYSKIND_SysKind {0:04X}".format(self.syskind))

        # PROJECTLCID Record
        # Specifies the VBA project's LCID.
        projectlcid_id = struct.unpack("<H", dir_stream.read(2))[0]
        self.check_value('PROJECTLCID_Id', 0x0002, projectlcid_id)
        projectlcid_size = struct.unpack("<L", dir_stream.read(4))[0]
        self.check_value('PROJECTLCID_Size', 0x0004, projectlcid_size)
        # Lcid (4 bytes): An unsigned integer that specifies the LCID value for the VBA project. MUST be 0x00000409.
        self.lcid = struct.unpack("<L", dir_stream.read(4))[0]
        self.check_value('PROJECTLCID_Lcid', 0x409, self.lcid)

        # PROJECTLCIDINVOKE Record
        # Specifies an LCID value used for Invoke calls on an Automation server as specified in [MS-OAUT] section 3.1.4.4.
        projectlcidinvoke_id = struct.unpack("<H", dir_stream.read(2))[0]
        self.check_value('PROJECTLCIDINVOKE_Id', 0x0014, projectlcidinvoke_id)
        projectlcidinvoke_size = struct.unpack("<L", dir_stream.read(4))[0]
        self.check_value('PROJECTLCIDINVOKE_Size',
                         0x0004, projectlcidinvoke_size)
        # LcidInvoke (4 bytes): An unsigned integer that specifies the LCID value used for Invoke calls. MUST be 0x00000409.
        self.lcidinvoke = struct.unpack("<L", dir_stream.read(4))[0]
        self.check_value('PROJECTLCIDINVOKE_LcidInvoke',
                         0x409, self.lcidinvoke)

        # PROJECTCODEPAGE Record
        # Specifies the VBA project's code page.
        projectcodepage_id = struct.unpack("<H", dir_stream.read(2))[0]
        self.check_value('PROJECTCODEPAGE_Id', 0x0003, projectcodepage_id)
        projectcodepage_size = struct.unpack("<L", dir_stream.read(4))[0]
        self.check_value('PROJECTCODEPAGE_Size', 0x0002, projectcodepage_size)
        self.codepage = struct.unpack("<H", dir_stream.read(2))[0]
        self.codepage_name = codepages.get_codepage_name(self.codepage)
        log.debug('Project Code Page: %r - %s' %
                  (self.codepage, self.codepage_name))
        self.codec = codepages.codepage2codec(self.codepage)
        log.debug('Python codec corresponding to code page %d: %s' %
                  (self.codepage, self.codec))

        # PROJECTNAME Record
        # Specifies a unique VBA identifier as the name of the VBA project.
        projectname_id = struct.unpack("<H", dir_stream.read(2))[0]
        self.check_value('PROJECTNAME_Id', 0x0004, projectname_id)
        sizeof_projectname = struct.unpack("<L", dir_stream.read(4))[0]
        log.debug('Project name size: %d bytes' % sizeof_projectname)
        if sizeof_projectname < 1 or sizeof_projectname > 128:
            # TODO: raise an actual error? What is MS Office's behaviour?
            log.error(
                "PROJECTNAME_SizeOfProjectName value not in range [1-128]: {0}".format(sizeof_projectname))
        projectname_bytes = dir_stream.read(sizeof_projectname)
        self.projectname = self.decode_bytes(projectname_bytes)

        # PROJECTDOCSTRING Record
        # Specifies the description for the VBA project.
        projectdocstring_id = struct.unpack("<H", dir_stream.read(2))[0]
        self.check_value('PROJECTDOCSTRING_Id', 0x0005, projectdocstring_id)
        projectdocstring_sizeof_docstring = struct.unpack(
            "<L", dir_stream.read(4))[0]
        if projectdocstring_sizeof_docstring > 2000:
            log.error(
                "PROJECTDOCSTRING_SizeOfDocString value not in range: {0}".format(projectdocstring_sizeof_docstring))
        # DocString (variable): An array of SizeOfDocString bytes that specifies the description for the VBA project.
        # MUST contain MBCS characters encoded using the code page specified in PROJECTCODEPAGE (section 2.3.4.2.1.4).
        # MUST NOT contain null characters.
        docstring_bytes = dir_stream.read(projectdocstring_sizeof_docstring)
        self.docstring = self.decode_bytes(docstring_bytes)
        projectdocstring_reserved = struct.unpack("<H", dir_stream.read(2))[0]
        self.check_value('PROJECTDOCSTRING_Reserved',
                         0x0040, projectdocstring_reserved)
        projectdocstring_sizeof_docstring_unicode = struct.unpack(
            "<L", dir_stream.read(4))[0]
        if projectdocstring_sizeof_docstring_unicode % 2 != 0:
            log.error("PROJECTDOCSTRING_SizeOfDocStringUnicode is not even")
        # DocStringUnicode (variable): An array of SizeOfDocStringUnicode bytes that specifies the description for the
        # VBA project. MUST contain UTF-16 characters. MUST NOT contain null characters.
        # MUST contain the UTF-16 encoding of DocString.
        docstring_unicode_bytes = dir_stream.read(
            projectdocstring_sizeof_docstring_unicode)
        self.docstring_unicode = docstring_unicode_bytes.decode(
            'utf16', errors='replace')

        # PROJECTHELPFILEPATH Record - MS-OVBA 2.3.4.2.1.7
        projecthelpfilepath_id = struct.unpack("<H", dir_stream.read(2))[0]
        self.check_value('PROJECTHELPFILEPATH_Id',
                         0x0006, projecthelpfilepath_id)
        projecthelpfilepath_sizeof_helpfile1 = struct.unpack(
            "<L", dir_stream.read(4))[0]
        if projecthelpfilepath_sizeof_helpfile1 > 260:
            log.error(
                "PROJECTHELPFILEPATH_SizeOfHelpFile1 value not in range: {0}".format(projecthelpfilepath_sizeof_helpfile1))
        projecthelpfilepath_helpfile1 = dir_stream.read(
            projecthelpfilepath_sizeof_helpfile1)
        projecthelpfilepath_reserved = struct.unpack(
            "<H", dir_stream.read(2))[0]
        self.check_value('PROJECTHELPFILEPATH_Reserved',
                         0x003D, projecthelpfilepath_reserved)
        projecthelpfilepath_sizeof_helpfile2 = struct.unpack(
            "<L", dir_stream.read(4))[0]
        if projecthelpfilepath_sizeof_helpfile2 != projecthelpfilepath_sizeof_helpfile1:
            log.error(
                "PROJECTHELPFILEPATH_SizeOfHelpFile1 does not equal PROJECTHELPFILEPATH_SizeOfHelpFile2")
        projecthelpfilepath_helpfile2 = dir_stream.read(
            projecthelpfilepath_sizeof_helpfile2)
        if projecthelpfilepath_helpfile2 != projecthelpfilepath_helpfile1:
            log.error(
                "PROJECTHELPFILEPATH_HelpFile1 does not equal PROJECTHELPFILEPATH_HelpFile2")

        # PROJECTHELPCONTEXT Record
        projecthelpcontext_id = struct.unpack("<H", dir_stream.read(2))[0]
        self.check_value('PROJECTHELPCONTEXT_Id',
                         0x0007, projecthelpcontext_id)
        projecthelpcontext_size = struct.unpack("<L", dir_stream.read(4))[0]
        self.check_value('PROJECTHELPCONTEXT_Size',
                         0x0004, projecthelpcontext_size)
        projecthelpcontext_helpcontext = struct.unpack(
            "<L", dir_stream.read(4))[0]
        unused = projecthelpcontext_helpcontext

        # PROJECTLIBFLAGS Record
        projectlibflags_id = struct.unpack("<H", dir_stream.read(2))[0]
        self.check_value('PROJECTLIBFLAGS_Id', 0x0008, projectlibflags_id)
        projectlibflags_size = struct.unpack("<L", dir_stream.read(4))[0]
        self.check_value('PROJECTLIBFLAGS_Size', 0x0004, projectlibflags_size)
        projectlibflags_projectlibflags = struct.unpack(
            "<L", dir_stream.read(4))[0]
        self.check_value('PROJECTLIBFLAGS_ProjectLibFlags',
                         0x0000, projectlibflags_projectlibflags)

        # PROJECTVERSION Record
        projectversion_id = struct.unpack("<H", dir_stream.read(2))[0]
        self.check_value('PROJECTVERSION_Id', 0x0009, projectversion_id)
        projectversion_reserved = struct.unpack("<L", dir_stream.read(4))[0]
        self.check_value('PROJECTVERSION_Reserved',
                         0x0004, projectversion_reserved)
        projectversion_versionmajor = struct.unpack(
            "<L", dir_stream.read(4))[0]
        projectversion_versionminor = struct.unpack(
            "<H", dir_stream.read(2))[0]
        unused = projectversion_versionmajor
        unused = projectversion_versionminor

        # PROJECTCONSTANTS Record
        projectconstants_id = struct.unpack("<H", dir_stream.read(2))[0]
        self.check_value('PROJECTCONSTANTS_Id', 0x000C, projectconstants_id)
        projectconstants_sizeof_constants = struct.unpack(
            "<L", dir_stream.read(4))[0]
        if projectconstants_sizeof_constants > 1015:
            log.error(
                "PROJECTCONSTANTS_SizeOfConstants value not in range: {0}".format(projectconstants_sizeof_constants))
        projectconstants_constants = dir_stream.read(
            projectconstants_sizeof_constants)
        projectconstants_reserved = struct.unpack("<H", dir_stream.read(2))[0]
        self.check_value('PROJECTCONSTANTS_Reserved',
                         0x003C, projectconstants_reserved)
        projectconstants_sizeof_constants_unicode = struct.unpack(
            "<L", dir_stream.read(4))[0]
        if projectconstants_sizeof_constants_unicode % 2 != 0:
            log.error("PROJECTCONSTANTS_SizeOfConstantsUnicode is not even")
        projectconstants_constants_unicode = dir_stream.read(
            projectconstants_sizeof_constants_unicode)
        unused = projectconstants_constants
        unused = projectconstants_constants_unicode

        # array of REFERENCE records
        # Specifies a reference to an Automation type library or VBA project.
        check = None
        while True:
            check = struct.unpack("<H", dir_stream.read(2))[0]
            log.debug("reference type = {0:04X}".format(check))
            if check == 0x000F:
                break

            if check == 0x0016:
                # REFERENCENAME
                # Specifies the name of a referenced VBA project or Automation type library.
                reference_id = check
                reference_sizeof_name = struct.unpack(
                    "<L", dir_stream.read(4))[0]
                reference_name = dir_stream.read(reference_sizeof_name)
                log.debug('REFERENCE name: %s' %
                          self.decode_bytes(reference_name))
                reference_reserved = struct.unpack("<H", dir_stream.read(2))[0]
                # According to [MS-OVBA] 2.3.4.2.2.2 REFERENCENAME Record:
                # "Reserved (2 bytes): MUST be 0x003E. MUST be ignored."
                # So let's ignore it, otherwise it crashes on some files (issue #132)
                # PR #135 by @c1fe:
                # contrary to the specification I think that the unicode name
                # is optional. if reference_reserved is not 0x003E I think it
                # is actually the start of another REFERENCE record
                # at least when projectsyskind_syskind == 0x02 (Macintosh)
                if reference_reserved == 0x003E:
                    # if reference_reserved not in (0x003E, 0x000D):
                    #    raise UnexpectedDataError(dir_path, 'REFERENCE_Reserved',
                    #                              0x0003E, reference_reserved)
                    reference_sizeof_name_unicode = struct.unpack(
                        "<L", dir_stream.read(4))[0]
                    reference_name_unicode = dir_stream.read(
                        reference_sizeof_name_unicode)
                    unused = reference_id
                    unused = reference_name
                    unused = reference_name_unicode
                    continue
                else:
                    check = reference_reserved
                    log.debug("reference type = {0:04X}".format(check))

            if check == 0x0033:
                # REFERENCEORIGINAL (followed by REFERENCECONTROL)
                # Specifies the identifier of the Automation type library the containing REFERENCECONTROL's
                # (section 2.3.4.2.2.3) twiddled type library was generated from.
                referenceoriginal_id = check
                referenceoriginal_sizeof_libidoriginal = struct.unpack(
                    "<L", dir_stream.read(4))[0]
                referenceoriginal_libidoriginal = dir_stream.read(
                    referenceoriginal_sizeof_libidoriginal)
                log.debug('REFERENCE original lib id: %s' %
                          self.decode_bytes(referenceoriginal_libidoriginal))
                unused = referenceoriginal_id
                unused = referenceoriginal_libidoriginal
                continue

            if check == 0x002F:
                # REFERENCECONTROL
                # Specifies a reference to a twiddled type library and its extended type library.
                referencecontrol_id = check
                referencecontrol_sizetwiddled = struct.unpack(
                    "<L", dir_stream.read(4))[0]  # ignore
                referencecontrol_sizeof_libidtwiddled = struct.unpack(
                    "<L", dir_stream.read(4))[0]
                referencecontrol_libidtwiddled = dir_stream.read(
                    referencecontrol_sizeof_libidtwiddled)
                log.debug('REFERENCE control twiddled lib id: %s' %
                          self.decode_bytes(referencecontrol_libidtwiddled))
                referencecontrol_reserved1 = struct.unpack(
                    "<L", dir_stream.read(4))[0]  # ignore
                self.check_value('REFERENCECONTROL_Reserved1',
                                 0x0000, referencecontrol_reserved1)
                referencecontrol_reserved2 = struct.unpack(
                    "<H", dir_stream.read(2))[0]  # ignore
                self.check_value('REFERENCECONTROL_Reserved2',
                                 0x0000, referencecontrol_reserved2)
                unused = referencecontrol_id
                unused = referencecontrol_sizetwiddled
                unused = referencecontrol_libidtwiddled
                # optional field
                check2 = struct.unpack("<H", dir_stream.read(2))[0]
                if check2 == 0x0016:
                    referencecontrol_namerecordextended_id = check
                    referencecontrol_namerecordextended_sizeof_name = struct.unpack(
                        "<L", dir_stream.read(4))[0]
                    referencecontrol_namerecordextended_name = dir_stream.read(
                        referencecontrol_namerecordextended_sizeof_name)
                    log.debug('REFERENCE control name record extended: %s' %
                              self.decode_bytes(referencecontrol_namerecordextended_name))
                    referencecontrol_namerecordextended_reserved = struct.unpack(
                        "<H", dir_stream.read(2))[0]
                    if referencecontrol_namerecordextended_reserved == 0x003E:
                        referencecontrol_namerecordextended_sizeof_name_unicode = struct.unpack(
                            "<L", dir_stream.read(4))[0]
                        referencecontrol_namerecordextended_name_unicode = dir_stream.read(
                            referencecontrol_namerecordextended_sizeof_name_unicode)
                        referencecontrol_reserved3 = struct.unpack(
                            "<H", dir_stream.read(2))[0]
                        unused = referencecontrol_namerecordextended_id
                        unused = referencecontrol_namerecordextended_name
                        unused = referencecontrol_namerecordextended_name_unicode
                    else:
                        referencecontrol_reserved3 = referencecontrol_namerecordextended_reserved
                else:
                    referencecontrol_reserved3 = check2

                self.check_value('REFERENCECONTROL_Reserved3',
                                 0x0030, referencecontrol_reserved3)
                referencecontrol_sizeextended = struct.unpack(
                    "<L", dir_stream.read(4))[0]
                referencecontrol_sizeof_libidextended = struct.unpack(
                    "<L", dir_stream.read(4))[0]
                referencecontrol_libidextended = dir_stream.read(
                    referencecontrol_sizeof_libidextended)
                referencecontrol_reserved4 = struct.unpack(
                    "<L", dir_stream.read(4))[0]
                referencecontrol_reserved5 = struct.unpack(
                    "<H", dir_stream.read(2))[0]
                referencecontrol_originaltypelib = dir_stream.read(16)
                referencecontrol_cookie = struct.unpack(
                    "<L", dir_stream.read(4))[0]
                unused = referencecontrol_sizeextended
                unused = referencecontrol_libidextended
                unused = referencecontrol_reserved4
                unused = referencecontrol_reserved5
                unused = referencecontrol_originaltypelib
                unused = referencecontrol_cookie
                continue

            if check == 0x000D:
                # REFERENCEREGISTERED
                # Specifies a reference to an Automation type library.
                referenceregistered_id = check
                referenceregistered_size = struct.unpack(
                    "<L", dir_stream.read(4))[0]
                referenceregistered_sizeof_libid = struct.unpack(
                    "<L", dir_stream.read(4))[0]
                referenceregistered_libid = dir_stream.read(
                    referenceregistered_sizeof_libid)
                log.debug('REFERENCE registered lib id: %s' %
                          self.decode_bytes(referenceregistered_libid))
                referenceregistered_reserved1 = struct.unpack(
                    "<L", dir_stream.read(4))[0]
                self.check_value('REFERENCEREGISTERED_Reserved1',
                                 0x0000, referenceregistered_reserved1)
                referenceregistered_reserved2 = struct.unpack(
                    "<H", dir_stream.read(2))[0]
                self.check_value('REFERENCEREGISTERED_Reserved2',
                                 0x0000, referenceregistered_reserved2)
                unused = referenceregistered_id
                unused = referenceregistered_size
                unused = referenceregistered_libid
                continue

            if check == 0x000E:
                # REFERENCEPROJECT
                # Specifies a reference to an external VBA project.
                referenceproject_id = check
                referenceproject_size = struct.unpack(
                    "<L", dir_stream.read(4))[0]
                referenceproject_sizeof_libidabsolute = struct.unpack(
                    "<L", dir_stream.read(4))[0]
                referenceproject_libidabsolute = dir_stream.read(
                    referenceproject_sizeof_libidabsolute)
                log.debug('REFERENCE project lib id absolute: %s' %
                          self.decode_bytes(referenceproject_libidabsolute))
                referenceproject_sizeof_libidrelative = struct.unpack(
                    "<L", dir_stream.read(4))[0]
                referenceproject_libidrelative = dir_stream.read(
                    referenceproject_sizeof_libidrelative)
                log.debug('REFERENCE project lib id relative: %s' %
                          self.decode_bytes(referenceproject_libidrelative))
                referenceproject_majorversion = struct.unpack(
                    "<L", dir_stream.read(4))[0]
                referenceproject_minorversion = struct.unpack(
                    "<H", dir_stream.read(2))[0]
                unused = referenceproject_id
                unused = referenceproject_size
                unused = referenceproject_libidabsolute
                unused = referenceproject_libidrelative
                unused = referenceproject_majorversion
                unused = referenceproject_minorversion
                continue

            log.error('invalid or unknown check Id {0:04X}'.format(check))
            # raise an exception instead of stopping abruptly (issue #180)
            raise UnexpectedDataError(
                dir_path, 'reference type', (0x0F, 0x16, 0x33, 0x2F, 0x0D, 0x0E), check)
            # sys.exit(0)

    def check_value(self, name, expected, value):
        if expected != value:
            if self.relaxed:
                log.error("invalid value for {0} expected {1:04X} got {2:04X}"
                          .format(name, expected, value))
            else:
                raise UnexpectedDataError(self.dir_path, name, expected, value)

    def parse_project_stream(self):
        """
        Parse the PROJECT stream from the VBA project
        :return:
        """
        # Open the PROJECT stream:
        # reference: [MS-OVBA] 2.3.1 PROJECT Stream
        project_stream = self.ole.openstream(self.project_path)

        # sample content of the PROJECT stream:

        # ID="{5312AC8A-349D-4950-BDD0-49BE3C4DD0F0}"
        # Document=ThisDocument/&H00000000
        # Module=NewMacros
        # Name="Project"
        # HelpContextID="0"
        # VersionCompatible32="393222000"
        # CMG="F1F301E705E705E705E705"
        # DPB="8F8D7FE3831F2020202020"
        # GC="2D2FDD81E51EE61EE6E1"
        ##
        # [Host Extender Info]
        # &H00000001={3832D640-CF90-11CF-8E43-00A0C911005A};VBE;&H00000000
        # &H00000002={000209F2-0000-0000-C000-000000000046};Word8.0;&H00000000
        ##
        # [Workspace]
        # ThisDocument=22, 29, 339, 477, Z
        # NewMacros=-4, 42, 832, 510, C

        self.module_ext = {}

        for line in project_stream:
            line = self.decode_bytes(line)
            log.debug('PROJECT: %r' % line)
            line = line.strip()
            if '=' in line:
                # split line at the 1st equal sign:
                name, value = line.split('=', 1)
                # looking for code modules
                # add the code module as a key in the dictionary
                # the value will be the extension needed later
                # The value is converted to lowercase, to allow case-insensitive matching (issue #3)
                value = value.lower()
                if name == 'Document':
                    # split value at the 1st slash, keep 1st part:
                    value = value.split('/', 1)[0]
                    self.module_ext[value] = CLASS_EXTENSION
                elif name == 'Module':
                    self.module_ext[value] = MODULE_EXTENSION
                elif name == 'Class':
                    self.module_ext[value] = CLASS_EXTENSION
                elif name == 'BaseClass':
                    self.module_ext[value] = FORM_EXTENSION

    def parse_modules(self):
        dir_stream = self.dir_stream
        # projectmodules_id has already been read by the previous loop = 0x000F
        # projectmodules_id = check  #struct.unpack("<H", dir_stream.read(2))[0]
        # self.check_value('PROJECTMODULES_Id', 0x000F, projectmodules_id)
        projectmodules_size = struct.unpack("<L", dir_stream.read(4))[0]
        self.check_value('PROJECTMODULES_Size', 0x0002, projectmodules_size)
        self.modules_count = struct.unpack("<H", dir_stream.read(2))[0]
        _id = struct.unpack("<H", dir_stream.read(2))[0]
        self.check_value('PROJECTMODULES_ProjectCookieRecord_Id', 0x0013, _id)
        size = struct.unpack("<L", dir_stream.read(4))[0]
        self.check_value(
            'PROJECTMODULES_ProjectCookieRecord_Size', 0x0002, size)
        projectcookierecord_cookie = struct.unpack("<H", dir_stream.read(2))[0]
        unused = projectcookierecord_cookie

        log.debug("parsing {0} modules".format(self.modules_count))
        for module_index in range(0, self.modules_count):
            module = VBA_Module(self, self.dir_stream,
                                module_index=module_index)
            self.modules.append(module)
            yield (module.code_path, module.filename_str, module.code_str)
        _ = unused   # make pylint happy: now variable "unused" is being used ;-)
        return

    def decode_bytes(self, bytes_string, errors='replace'):
        """
        Decode a bytes string to a unicode string, using the project code page
        :param bytes_string: bytes, bytes string to be decoded
        :param errors: str, mode to handle unicode conversion errors
        :return: str/unicode, decoded string
        """
        return bytes_string.decode(self.codec, errors=errors)


class VBA_RUN(object):
    def __init__(self):
        self.ole = None

    def is_ole(self, content):
        return olefile.isOleFile(content)

    def check_vba_stream(self, vba_root, stream_path):
        full_path = vba_root + stream_path
        if self.ole.exists(full_path) and self.ole.get_type(full_path) == olefile.STGTY_STREAM:
            return full_path
        else:
            return False

    def get_vba_projects(self):
        # start with an empty list:
        vba_projects = []
        # Look for any storage containing those storage/streams:
        for storage in self.ole.listdir(streams=False, storages=True):
            # Look for a storage ending with "VBA":
            if storage[-1].upper() == 'VBA':
                vba_root = '/'.join(storage[:-1])
                # Add a trailing slash to vba_root, unless it is the root of the OLE file:
                # (used later to append all the child streams/storages)
                if vba_root != '':
                    vba_root += '/'

                # Check if the VBA root storage also contains a PROJECT stream:
                project_path = self.check_vba_stream(vba_root, 'PROJECT')
                if not project_path:
                    continue
                # Check if the VBA root storage also contains a VBA/_VBA_PROJECT stream:
                vba_project_path = self.check_vba_stream(
                    vba_root, 'VBA/_VBA_PROJECT')
                if not vba_project_path:
                    continue
                # Check if the VBA root storage also contains a VBA/dir stream:
                dir_path = self.check_vba_stream(vba_root, 'VBA/dir')
                if not dir_path:
                    continue
                # Now we are pretty sure it is a VBA project structure
                # append the results to the list as a tuple for later use:
                vba_projects.append((vba_root, project_path, dir_path))
        return vba_projects

    def _extract_vba(self, vba_root, project_path, dir_path, relaxed=True):
        """
        Extract VBA macros from an OleFileIO object.
        Internal function, do not call directly.

        vba_root: path to the VBA root storage, containing the VBA storage and the PROJECT stream
        vba_project: path to the PROJECT stream
        :param relaxed: If True, only create info/debug log entry if data is not as expected
                        (e.g. opening substream fails); if False, raise an error in this case
        This is a generator, yielding (stream path, VBA filename, VBA source code) for each VBA code stream
        """

        project = VBA_Project(self.ole, vba_root,
                              project_path, dir_path, relaxed)
        project.parse_project_stream()

        for code_path, filename, code_data in project.parse_modules():
            yield (code_path, filename, code_data)

    def parse(self, file_content):
        if olefile.isOleFile(file_content):
            self.ole = olefile.OleFileIO(file_content)
            vba_projects = self.get_vba_projects()
            for vba_root, project_path, dir_path in vba_projects:
                # extract all VBA macros from that VBA root storage:
                # The function _extract_vba may fail on some files (issue #132)
                try:
                    for stream_path, vba_filename, vba_code in self._extract_vba(vba_root, project_path, dir_path, True):
                        yield {"stream_path": stream_path, "vba_filename": vba_filename, "vba_code": vba_code.encode()}
                except Exception as e:
                    log.exception('Error in _extract_vba')



# -*- coding: utf-8 -*-
# @Time    : 2020/12/23 01:57:47 
# @Author  : ddvv
# @Site    : https://ddvvmmzz.github.io
# @File    : exceptions.py 
# @Software: Visual Studio Code


class CriticalError(Exception):
    """critical error."""


class DependencyError(CriticalError):
    """Missing dependency error."""


class ProcessingError(CriticalError):
    """Error in processor module."""



# -*- coding: utf-8 -*-
# @Time    : 2020/12/20 16:25:45
# @Author  : ddvv
# @Site    : https://ddvvmmzz.github.io
# @File    : html_infos.py
# @Software: Visual Studio Code


from data_lib.mmpi.common.abstracts import Signature
from mmpi.data.white import check_fitler_domain


class ProbeDetection(Signature):
    authors = ["ddvv"]
    sig_type = 'html'
    name = "probe_detection"
    severity = 6
    description = "Mail Probe Detection"

    def on_complete(self):
        results = self.get_results()
        for result in results:
            if result.get('type', '') == self.sig_type:
                infos = result.get('value', {}).get('infos', [])
                complex_number = len(infos)
                if complex_number < 5:
                    for info in infos:
                        if info.get('type', '') == 'img':
                            data = info.get('data', {})
                            src = data.get('src', '')
                            # suffix = src.split('.')[-1]
                            # if suffix not in ['gif', 'png', 'jpeg', 'jpg', 'pic'] and not check_fitler_domain(src):
                            if not check_fitler_domain(src):
                                width = data.get('width', 9999)
                                height = data.get('height', 9999)
                                if width + height <= 2:
                                        self.mark(type="html", data=data.get('src'), tag=self.name)
                    return self.has_marks()
        return None

class PhishingDetection(Signature):
    authors = ["ddvv"]
    sig_type = 'html'
    name = "phishing_detection"
    severity = 6
    description = "Mail Phishing Detection"

    def on_complete(self):
        results = self.get_results()
        marks = list()
        for result in results:
            if result.get('type', '') == self.sig_type:
                infos = result.get('value', {}).get('infos', [])
                for info in infos:
                    if info.get('type', '') == 'hyperlink':
                        data = info.get('data', {})
                        src = data.get('src')
                        if not check_fitler_domain(src):
                            word = data.get('word')
                            if word not in src:
                                last_word = src.replace('/', '')[-4:].lower()
                                url_param = src.split('&')
                                url_path = src.split('/')
                                domains = word.replace('www.', '')
                                complex_number = len(infos)
                                if (last_word != word[-4:].lower() or complex_number < 5) and domains not in src and len(url_param) < 4 and (last_word == '.php' or len(url_path) < 12):
                                    # self.mark(type="html", data=data.get('src'), tag=self.name)
                                    marks.append({"type": "html", "data": data.get("src"), "tag": self.name})
        if marks and len(marks) < 5:
            for mark in marks:
                self.mark(**mark)
            return self.has_marks()
        else:
            return None

class SPAMDetection(Signature):
    authors = ["ddvv"]
    sig_type = 'html'
    name = "spam_detection"
    severity = 3
    description = "SPAM Detection"

    def on_complete(self):
        results = self.get_results()
        for result in results:
            if result.get('type', '') == self.sig_type:
                infos = result.get('value', {}).get('infos', [])
                complex_number = len(infos)
                if complex_number > 9:
                    self.mark(type="html", tag=self.name)
                    return self.has_marks()
        return None



# -*- coding: utf-8 -*-
# @Time    : 2020/12/23 01:58:57 
# @Author  : ddvv
# @Site    : https://ddvvmmzz.github.io
# @File    : compat.py 
# @Software: Visual Studio Code


def enumerate_signatures(dirpath, submodule, g, attributes):
    """In the new Cuckoo package, Signatures are no longer accessed
    under the modules module."""
    try:
        from mmpi.core.plugins import enumerate_plugins
        from mmpi.common.abstracts import Signature

        return enumerate_plugins(
            dirpath, "signatures.%s" % submodule,
            g, Signature, attributes
        )
    except ImportError:
        from mmpi.core.plugins import enumerate_plugins
        from mmpi.common.abstracts import Signature

        return enumerate_plugins(
            dirpath, "modules.signatures.%s" % submodule,
            g, Signature, attributes
        )



# -*- coding: utf-8 -*-
# @Time    : 2020/12/18 22:02:31 
# @Author  : ddvv
# @Site    : https://ddvvmmzz.github.io
# @File    : main.py 
# @Software: Visual Studio Code


import json
from string import ascii_letters
import sys
from mmpi.core.plugins import RunProcessing, RunSignatures
from mmpi.common.filetypes import get_support_file_type
from mmpi.core.startup import init_modules
from mmpi.misc import load_signatures


class mmpi(object):

    def __init__(self):
        load_signatures()
        init_modules()
        self.results = list()
        self.reports = dict()

    def get_results(self):
        return self.results

    def process(self, data):
        file_type = get_support_file_type(data.get('file_type'))
        results = RunProcessing(data).run()
        if results:
            self.results.append(results)
            report = results.get('value').get('report', {})
            if report:
                self.reports.update(report)
            if file_type != 'other':
                datas = results.get('value', {}).get('datas', [])
                for data in datas:
                    self.process(data)

    def signatures(self, data):
        RunSignatures(data).run()

    def parse(self, filename):
        self.results = list()
        self.reports = dict()
        with open(filename, 'rb') as f:
            data = {'content': f.read(), 'file_type': 'eml'}
        if data.get('content', b''):
            self.process(data)
            self.signatures(self.results)
            self.reports['signatures'] = self.get_signatures()

    def get_report(self):
        return self.reports

    def get_signatures(self):
        signatures = list()
        for result in self.results:
            file_type = result.get('type')
            if file_type == 'signatures':
                infos = result.get('value', {}).get('infos')
                for info in infos:
                    marks = info.get('marks', [])
                    marks = [dict(t) for t in set([tuple(mark.items()) for mark in marks])]
                    info['marks'] = marks
                    signatures.append(info)

        return signatures


def main():
    mmpi_ins = mmpi()
    mmpi_ins.parse(sys.argv[1])
    report = mmpi_ins.get_report()
    print(json.dumps(report, indent=4, ensure_ascii=False))



# -*- coding: utf-8 -*-
# @Time    : 2020/12/30 00:20:49 
# @Author  : ddvv
# @Site    : https://ddvvmmzz.github.io
# @File    : rtfextract.py 
# @Software: Visual Studio Code


from __future__ import print_function

"""
rtfobj.py

rtfobj is a Python module to extract embedded objects from RTF files, such as
OLE ojects. It can be used as a Python library or a command-line tool.

Usage: rtfobj.py <file.rtf>

rtfobj project website: http://www.decalage.info/python/rtfobj

rtfobj is part of the python-oletools package:
http://www.decalage.info/python/oletools
"""

#=== LICENSE =================================================================

# rtfobj is copyright (c) 2012-2019, Philippe Lagadec (http://www.decalage.info)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#------------------------------------------------------------------------------
# CHANGELOG:
# 2012-11-09 v0.01 PL: - first version
# 2013-04-02 v0.02 PL: - fixed bug in main
# 2015-12-09 v0.03 PL: - configurable logging, CLI options
#                      - extract OLE 1.0 objects
#                      - extract files from OLE Package objects
# 2016-04-01 v0.04 PL: - fixed logging output to use stdout instead of stderr
# 2016-04-07 v0.45 PL: - improved parsing to handle some malware tricks
# 2016-05-06 v0.47 TJ: - added option -d to set the output directory
#                        (contribution by Thomas Jarosch)
#                  TJ: - sanitize filenames to avoid special characters
# 2016-05-29       PL: - improved parsing, fixed issue #42
# 2016-07-13 v0.50 PL: - new RtfParser and RtfObjParser classes
# 2016-07-18       SL: - added Python 3.5 support
# 2016-07-19       PL: - fixed Python 2.6-2.7 support
# 2016-07-30       PL: - new API with class RtfObject
#                      - backward-compatible API rtf_iter_objects (fixed issue #70)
# 2016-07-31       PL: - table output with tablestream
# 2016-08-01       PL: - detect executable filenames in OLE Package
# 2016-08-08       PL: - added option -s to save objects to files
# 2016-08-09       PL: - fixed issue #78, improved regex
# 2016-09-06       PL: - fixed issue #83, backward compatible API
# 2016-11-17 v0.51 PL: - updated call to oleobj.OleNativeStream
# 2017-03-12       PL: - fixed imports for Python 2+3
#                      - fixed hex decoding bug in RtfObjParser (issue #103)
# 2017-03-29       PL: - fixed RtfParser to handle issue #152 (control word with
#                        long parameter)
# 2017-04-11       PL: - added detection of the OLE2Link vulnerability CVE-2017-0199
# 2017-05-04       PL: - fixed issue #164 to handle linked OLE objects
# 2017-06-08       PL: - fixed issue/PR #143: bin object with negative length
# 2017-06-29       PL: - temporary fix for issue #178
# 2017-07-14 v0.52 PL: - disabled logging of each control word (issue #184)
# 2017-07-24       PL: - fixed call to RtfParser._end_of_file (issue #185)
#                      - ignore optional space after \bin (issue #185)
# 2017-09-06       PL: - fixed issue #196: \pxe is not a destination
# 2018-01-11       CH: - speedup RTF parsing (PR #244)
# 2018-02-01      JRM: - fixed issue #251: \bin without argument
# 2018-04-09       PL: - fixed issue #280: OLE Package were not detected on Python 3
# 2018-03-24 v0.53 PL: - fixed issue #292: \margSz is a destination
# 2018-04-27       PL: - extract and display the CLSID of OLE objects
# 2018-04-30       PL: - handle "\'" obfuscation trick - issue #281
# 2018-05-10       PL: - fixed issues #303 #307: several destination cwords were incorrect
# 2018-05-17       PL: - fixed issue #273: bytes constants instead of str
# 2018-05-31 v0.53.1 PP: - fixed issue #316: whitespace after \bin on Python 3
# 2018-06-22 v0.53.2 PL: - fixed issue #327: added "\pnaiu" & "\pnaiud"
# 2018-09-11 v0.54 PL: - olefile is now a dependency
# 2019-07-08 v0.55 MM: - added URL carver for CVE-2017-0199 (Equation Editor) PR #460
#                      - added SCT to the list of executable file extensions PR #461
# 2019-12-16 v0.55.2 PL: - \rtf is not a destination control word (issue #522)
# 2019-12-17         PL: - fixed process_file to detect Equation class (issue #525)

__version__ = '0.55.2'

# ------------------------------------------------------------------------------
# TODO:
# - allow semicolon within hex, as found in  this sample:
#   http://contagiodump.blogspot.nl/2011/10/sep-28-cve-2010-3333-manuscript-with.html
# TODO: use OleObject and OleNativeStream in RtfObject instead of copying each attribute
# TODO: option -e <id> to extract an object, -e all for all objects
# TODO: option to choose which destinations to include (objdata by default)
# TODO: option to display SHA256 or MD5 hashes of objects in table


# === IMPORTS =================================================================

import re, os, sys, binascii, logging, optparse, hashlib
import os.path
from time import time

# IMPORTANT: it should be possible to run oletools directly as scripts
# in any directory without installing them with pip or setup.py.
# In that case, relative imports are NOT usable.
# And to enable Python 2+3 compatibility, we need to use absolute imports,
# so we add the oletools parent folder to sys.path (absolute+normalized path):
_thismodule_dir = os.path.normpath(os.path.abspath(os.path.dirname(__file__)))
# print('_thismodule_dir = %r' % _thismodule_dir)
_parent_dir = os.path.normpath(os.path.join(_thismodule_dir, '..'))
# print('_parent_dir = %r' % _thirdparty_dir)
if not _parent_dir in sys.path:
    sys.path.insert(0, _parent_dir)

import olefile
from mmpi.common import oleobj
from mmpi.common import clsid

# === LOGGING =================================================================

class NullHandler(logging.Handler):
    """
    Log Handler without output, to avoid printing messages if logging is not
    configured by the main application.
    Python 2.7 has logging.NullHandler, but this is necessary for 2.6:
    see https://docs.python.org/2.6/library/logging.html#configuring-logging-for-a-library
    """
    def emit(self, record):
        pass

def get_logger(name, level=logging.CRITICAL+1):
    """
    Create a suitable logger object for this module.
    The goal is not to change settings of the root logger, to avoid getting
    other modules' logs on the screen.
    If a logger exists with same name, reuse it. (Else it would have duplicate
    handlers and messages would be doubled.)
    The level is set to CRITICAL+1 by default, to avoid any logging.
    """
    # First, test if there is already a logger with the same name, else it
    # will generate duplicate messages (due to duplicate handlers):
    if name in logging.Logger.manager.loggerDict:
        #NOTE: another less intrusive but more "hackish" solution would be to
        # use getLogger then test if its effective level is not default.
        logger = logging.getLogger(name)
        # make sure level is OK:
        logger.setLevel(level)
        return logger
    # get a new logger:
    logger = logging.getLogger(name)
    # only add a NullHandler for this logger, it is up to the application
    # to configure its own logging:
    logger.addHandler(NullHandler())
    logger.setLevel(level)
    return logger

# a global logger object used for debugging:
log = get_logger('rtfobj')


#=== CONSTANTS=================================================================

# REGEX pattern to extract embedded OLE objects in hexadecimal format:

# alphanum digit: [0-9A-Fa-f]
HEX_DIGIT = b'[0-9A-Fa-f]'

# hex char = two alphanum digits: [0-9A-Fa-f]{2}
# HEX_CHAR = r'[0-9A-Fa-f]{2}'
# in fact MS Word allows whitespaces in between the hex digits!
# HEX_CHAR = r'[0-9A-Fa-f]\s*[0-9A-Fa-f]'
# Even worse, MS Word also allows ANY RTF-style tag {*} in between!!
# AND the tags can be nested...
#SINGLE_RTF_TAG = r'[{][^{}]*[}]'
# Actually RTF tags may contain braces escaped with backslash (\{ \}):
SINGLE_RTF_TAG = b'[{](?:\\\\.|[^{}\\\\])*[}]'

# Nested tags, two levels (because Python's re does not support nested matching):
# NESTED_RTF_TAG = r'[{](?:[^{}]|'+SINGLE_RTF_TAG+r')*[}]'
NESTED_RTF_TAG = b'[{](?:\\\\.|[^{}\\\\]|'+SINGLE_RTF_TAG+b')*[}]'

# AND it is also allowed to insert ANY control word or control symbol (ignored)
# According to Rich Text Format (RTF) Specification Version 1.9.1,
# section "Control Word":
# control word = \<ASCII Letter [a-zA-Z] Sequence max 32><Delimiter>
# delimiter = space, OR signed integer followed by any non-digit,
#             OR any character except letter and digit
# examples of valid control words:
# "\AnyThing " "\AnyThing123z" ""\AnyThing-456{" "\AnyThing{"
# control symbol = \<any char except letter or digit> (followed by anything)

ASCII_NAME = b'([a-zA-Z]{1,250})'

# using Python's re lookahead assumption:
# (?=...) Matches if ... matches next, but doesn't consume any of the string.
# This is called a lookahead assertion. For example, Isaac (?=Asimov) will
# match 'Isaac ' only if it's followed by 'Asimov'.

# TODO: Find the actual limit on the number of digits for Word
# SIGNED_INTEGER = r'(-?\d{1,250})'
SIGNED_INTEGER = b'(-?\\d+)'

# Note for issue #78: need to match "\A-" not followed by digits (or the end of string)
CONTROL_WORD = b'(?:\\\\' + ASCII_NAME + b'(?:' + SIGNED_INTEGER + b'(?=[^0-9])|(?=[^a-zA-Z0-9])|$))'

re_control_word = re.compile(CONTROL_WORD)

# Note for issue #78: need to match "\" followed by digit (any non-alpha)
CONTROL_SYMBOL = b'(?:\\\\[^a-zA-Z])'
re_control_symbol = re.compile(CONTROL_SYMBOL)

# Text that is not a control word/symbol or a group:
TEXT = b'[^{}\\\\]+'
re_text = re.compile(TEXT)

# ignored whitespaces and tags within a hex block:
IGNORED = b'(?:\\s|'+NESTED_RTF_TAG+b'|'+CONTROL_SYMBOL+b'|'+CONTROL_WORD+b')*'
#IGNORED = r'\s*'

# HEX_CHAR = HEX_DIGIT + IGNORED + HEX_DIGIT

# several hex chars, at least 4: (?:[0-9A-Fa-f]{2}){4,}
# + word boundaries
# HEX_CHARS_4orMORE = r'\b(?:' + HEX_CHAR + r'){4,}\b'
# at least 1 hex char:
# HEX_CHARS_1orMORE = r'(?:' + HEX_CHAR + r')+'
# at least 1 hex char, followed by whitespace or CR/LF:
# HEX_CHARS_1orMORE_WHITESPACES = r'(?:' + HEX_CHAR + r')+\s+'
# + word boundaries around hex block
# HEX_CHARS_1orMORE_WHITESPACES = r'\b(?:' + HEX_CHAR + r')+\b\s*'
# at least one block of hex and whitespace chars, followed by closing curly bracket:
# HEX_BLOCK_CURLY_BRACKET = r'(?:' + HEX_CHARS_1orMORE_WHITESPACES + r')+\}'
# PATTERN = r'(?:' + HEX_CHARS_1orMORE_WHITESPACES + r')*' + HEX_CHARS_1orMORE

#TODO PATTERN = r'\b(?:' + HEX_CHAR + IGNORED + r'){4,}\b'
# PATTERN = r'\b(?:' + HEX_CHAR + IGNORED + r'){4,}' #+ HEX_CHAR + r'\b'
PATTERN = b'\\b(?:' + HEX_DIGIT + IGNORED + b'){7,}' + HEX_DIGIT + b'\\b'

# at least 4 hex chars, followed by whitespace or CR/LF: (?:[0-9A-Fa-f]{2}){4,}\s*
# PATTERN = r'(?:(?:[0-9A-Fa-f]{2})+\s*)*(?:[0-9A-Fa-f]{2}){4,}'
# improved pattern, allowing semicolons within hex:
#PATTERN = r'(?:(?:[0-9A-Fa-f]{2})+\s*)*(?:[0-9A-Fa-f]{2}){4,}'

re_hexblock = re.compile(PATTERN)
re_embedded_tags = re.compile(IGNORED)
re_decimal = re.compile(b'\\d+')

re_delimiter = re.compile(b'[ \\t\\r\\n\\f\\v]')

DELIMITER = b'[ \\t\\r\\n\\f\\v]'
DELIMITERS_ZeroOrMore = b'[ \\t\\r\\n\\f\\v]*'
BACKSLASH_BIN = b'\\\\bin'
# According to my tests, Word accepts up to 250 digits (leading zeroes)
DECIMAL_GROUP = b'(\d{1,250})'

re_delims_bin_decimal = re.compile(DELIMITERS_ZeroOrMore + BACKSLASH_BIN
                                   + DECIMAL_GROUP + DELIMITER)
re_delim_hexblock = re.compile(DELIMITER + PATTERN)

# TODO: use a frozenset instead of a regex?
re_executable_extensions = re.compile(
    r"(?i)\.(BAT|CLASS|CMD|CPL|DLL|EXECOM|GADGET|HTA|INF|JAR|JS|JSE|LNK|MSC|MSI|MSP|PIF|PS1|PS1XML|PS2|PS2XML|PSC1|PSC2|REG|SCF|SCR|SCT|VB|VBE|VBS|WS|WSC|WSF|WSH)\b")

# Destination Control Words, according to MS RTF Specifications v1.9.1:
DESTINATION_CONTROL_WORDS = frozenset((
    b"aftncn", b"aftnsep", b"aftnsepc", b"annotation", b"atnauthor", b"atndate", b"atnid", b"atnparent", b"atnref",
    b"atrfend", b"atrfstart", b"author", b"background", b"bkmkend", b"bkmkstart", b"blipuid", b"buptim", b"category",
    b"colorschememapping", b"colortbl", b"comment", b"company", b"creatim", b"datafield", b"datastore", b"defchp", b"defpap",
    b"do", b"doccomm", b"docvar", b"dptxbxtext", b"ebcend", b"ebcstart", b"factoidname", b"falt", b"fchars", b"ffdeftext",
    b"ffentrymcr", b"ffexitmcr", b"ffformat", b"ffhelptext", b"ffl", b"ffname",b"ffstattext", b"field", b"file", b"filetbl",
    b"fldinst", b"fldrslt", b"fldtype", b"fontemb", b"fonttbl", b"footer", b"footerf", b"footerl",
    b"footerr", b"footnote", b"formfield", b"ftncn", b"ftnsep", b"ftnsepc", b"g", b"generator", b"gridtbl", b"header", b"headerf",
    b"headerl", b"headerr", b"hl", b"hlfr", b"hlinkbase", b"hlloc", b"hlsrc", b"hsv", b"info", b"keywords",
    b"latentstyles", b"lchars", b"levelnumbers", b"leveltext", b"lfolevel", b"linkval", b"list", b"listlevel", b"listname",
    b"listoverride", b"listoverridetable", b"listpicture", b"liststylename", b"listtable", b"listtext", b"lsdlockedexcept",
    b"macc", b"maccPr", b"mailmerge", b"malnScr", b"manager", b"margPr", b"mbar", b"mbarPr", b"mbaseJc", b"mbegChr",
    b"mborderBox", b"mborderBoxPr", b"mbox", b"mboxPr", b"mchr", b"mcount", b"mctrlPr", b"md", b"mdeg", b"mdegHide", b"mden",
    b"mdiff", b"mdPr", b"me", b"mendChr", b"meqArr", b"meqArrPr", b"mf", b"mfName", b"mfPr", b"mfunc", b"mfuncPr",b"mgroupChr",
    b"mgroupChrPr",b"mgrow", b"mhideBot", b"mhideLeft", b"mhideRight", b"mhideTop", b"mlim", b"mlimLoc", b"mlimLow",
    b"mlimLowPr", b"mlimUpp", b"mlimUppPr", b"mm", b"mmaddfieldname", b"mmathPict", b"mmaxDist", b"mmc",
    b"mmcJc", b"mmconnectstr", b"mmconnectstrdata", b"mmcPr", b"mmcs", b"mmdatasource", b"mmheadersource", b"mmmailsubject",
    b"mmodso", b"mmodsofilter", b"mmodsofldmpdata", b"mmodsomappedname", b"mmodsoname", b"mmodsorecipdata", b"mmodsosort",
    b"mmodsosrc", b"mmodsotable", b"mmodsoudl", b"mmodsoudldata", b"mmodsouniquetag", b"mmPr", b"mmquery", b"mmr", b"mnary",
    b"mnaryPr", b"mnoBreak", b"mnum", b"mobjDist", b"moMath", b"moMathPara", b"moMathParaPr", b"mopEmu", b"mphant", b"mphantPr",
    b"mplcHide", b"mpos", b"mr", b"mrad", b"mradPr", b"mrPr", b"msepChr", b"mshow", b"mshp", b"msPre", b"msPrePr", b"msSub",
    b"msSubPr", b"msSubSup", b"msSubSupPr",  b"msSup", b"msSupPr", b"mstrikeBLTR", b"mstrikeH", b"mstrikeTLBR", b"mstrikeV",
    b"msub", b"msubHide", b"msup", b"msupHide", b"mtransp", b"mtype", b"mvertJc", b"mvfmf", b"mvfml", b"mvtof", b"mvtol",
    b"mzeroAsc", b"mzeroDesc", b"mzeroWid", b"nesttableprops", b"nonesttables", b"objalias", b"objclass",
    b"objdata", b"object", b"objname", b"objsect", b"oldcprops", b"oldpprops", b"oldsprops", b"oldtprops",
    b"oleclsid", b"operator", b"panose", b"password", b"passwordhash", b"pgp", b"pgptbl", b"picprop", b"pict", b"pn", b"pnseclvl",
    b"pntext", b"pntxta", b"pntxtb", b"printim",
    b"propname", b"protend", b"protstart", b"protusertbl",
    b"result", b"revtbl", b"revtim",
    # \rtf should not be treated as a destination (issue #522)
    #b"rtf",
    b"rxe", b"shp", b"shpgrp", b"shpinst", b"shppict", b"shprslt", b"shptxt",
    b"sn", b"sp", b"staticval", b"stylesheet", b"subject", b"sv", b"svb", b"tc", b"template", b"themedata", b"title", b"txe", b"ud",
    b"upr", b"userprops", b"wgrffmtfilter", b"windowcaption", b"writereservation", b"writereservhash", b"xe", b"xform",
    b"xmlattrname", b"xmlattrvalue", b"xmlclose", b"xmlname", b"xmlnstbl", b"xmlopen",
    # added for issue #292: https://github.com/decalage2/oletools/issues/292
    b"margSz",
    # added for issue #327:
    b"pnaiu", b"pnaiud",

    # It seems \private should not be treated as a destination (issue #178)
    # Same for \pxe (issue #196)
    # b"private", b"pxe",
    # from issue #303: These destination control words can be treated as a "value" type.
    # They don't consume data so they won't change the state of the parser.
    # b"atnicn", b"atntime", b"fname", b"fontfile", b"htmltag", b"keycode", b"maln",
    # b"mhtmltag", b"mmath", b"mmathPr", b"nextfile", b"objtime", b"rsidtbl",
    ))


# some str methods on Python 2.x return characters,
# while the equivalent bytes methods return integers on Python 3.x:
if sys.version_info[0] <= 2:
    # Python 2.x - Characters (str)
    BACKSLASH = '\\'
    BRACE_OPEN = '{'
    BRACE_CLOSE = '}'
    UNICODE_TYPE = unicode
else:
    # Python 3.x - Integers
    BACKSLASH = ord('\\')
    BRACE_OPEN = ord('{')
    BRACE_CLOSE = ord('}')
    UNICODE_TYPE = str

RTF_MAGIC = b'\x7b\\rt'   # \x7b == b'{' but does not mess up auto-indent


#=== CLASSES =================================================================

class Destination(object):
    """
    Stores the data associated with a destination control word
    """
    def __init__(self, cword=None):
        self.cword = cword
        self.data = b''
        self.start = None
        self.end = None
        self.group_level = 0


class RtfParser(object):
    """
    Very simple but robust generic RTF parser, designed to handle
    malformed malicious RTF as MS Word does
    """

    def __init__(self, data):
        """
        RtfParser constructor.
        
        :param data: bytes object containing the RTF data to be parsed 
        """
        self.data = data
        self.index = 0
        self.size = len(data)
        self.group_level = 0
        # default destination for the document text:
        document_destination = Destination()
        self.destinations = [document_destination]
        self.current_destination = document_destination

    def parse(self):
        """
        Parse the RTF data
        
        :return: nothing
        """
        # Start at beginning of data
        self.index = 0
        start_time = time()
        last_report = start_time
        # Loop until the end
        while self.index < self.size:
            if time() - last_report > 15:     # report every 15s
                last_report = time()
            if self.data[self.index] == BRACE_OPEN:
                # Found an opening brace "{": Start of a group
                self._open_group()
                self.index += 1
                continue
            if self.data[self.index] == BRACE_CLOSE:
                # Found a closing brace "}": End of a group
                self._close_group()
                self.index += 1
                continue
            if self.data[self.index] == BACKSLASH:
                # Found a backslash "\": Start of a control word or control symbol
                # Use a regex to extract the control word name if present:
                # NOTE: the full length of the control word + its optional integer parameter
                # is limited by MS Word at 253 characters, so we have to run the regex
                # on a cropped string:
                data_cropped = self.data[self.index:self.index+254]
                # append a space so that the regex can check the following character:
                data_cropped += b' '
                # m = re_control_word.match(self.data, self.index, self.index+253)
                m = re_control_word.match(data_cropped)
                if m:
                    cword = m.group(1)
                    param = None
                    if len(m.groups()) > 1:
                        param = m.group(2)
                    # log.debug('control word at index %Xh - cword=%r param=%r  %r' % (self.index, cword, param, m.group()))
                    self._control_word(m, cword, param)
                    self.index += len(m.group())
                    # if it's \bin, call _bin after updating index
                    if cword == b'bin':
                        self._bin(m, param)
                    continue
                # Otherwise, it may be a control symbol:
                m = re_control_symbol.match(self.data, self.index)
                if m:
                    self.control_symbol(m)
                    self.index += len(m.group())
                    continue
            # Otherwise, this is plain text:
            # Use a regex to match all characters until the next brace or backslash:
            m = re_text.match(self.data, self.index)
            if m:
                self._text(m)
                self.index += len(m.group())
                continue
            raise RuntimeError('Should not have reached this point - index=%Xh' % self.index)
        # call _end_of_file to make sure all groups are closed properly
        self._end_of_file()


    def _open_group(self):
        self.group_level += 1
        #log.debug('{ Open Group at index %Xh - level=%d' % (self.index, self.group_level))
        # call user method AFTER increasing the level:
        self.open_group()

    def open_group(self):
        #log.debug('open group at index %Xh' % self.index)
        pass

    def _close_group(self):
        #log.debug('} Close Group at index %Xh - level=%d' % (self.index, self.group_level))
        # call user method BEFORE decreasing the level:
        self.close_group()
        # if the destination level is the same as the group level, close the destination:
        if self.group_level == self.current_destination.group_level:
            # log.debug('Current Destination %r level = %d => Close Destination' % (
            #     self.current_destination.cword, self.current_destination.group_level))
            self._close_destination()
        else:
            # log.debug('Current Destination %r level = %d => Continue with same Destination' % (
            #     self.current_destination.cword, self.current_destination.group_level))
            pass
        self.group_level -= 1
        # log.debug('Decreased group level to %d' % self.group_level)

    def close_group(self):
        #log.debug('close group at index %Xh' % self.index)
        pass

    def _open_destination(self, matchobject, cword):
        # if the current destination is at the same group level, close it first:
        if self.current_destination.group_level == self.group_level:
            self._close_destination()
        new_dest = Destination(cword)
        new_dest.group_level = self.group_level
        self.destinations.append(new_dest)
        self.current_destination = new_dest
        # start of the destination is right after the control word:
        new_dest.start = self.index + len(matchobject.group())
        # log.debug("Open Destination %r start=%Xh - level=%d" % (cword, new_dest.start, new_dest.group_level))
        # call the corresponding user method for additional processing:
        self.open_destination(self.current_destination)

    def open_destination(self, destination):
        pass

    def _close_destination(self):
        # log.debug("Close Destination %r end=%Xh - level=%d" % (self.current_destination.cword,
        #     self.index, self.current_destination.group_level))
        self.current_destination.end = self.index
        # call the corresponding user method for additional processing:
        self.close_destination(self.current_destination)
        if len(self.destinations)>0:
            # remove the current destination from the stack, and go back to the previous one:
            self.destinations.pop()
        if len(self.destinations) > 0:
            self.current_destination = self.destinations[-1]
        else:
            # log.debug('All destinations are closed, keeping the document destination open')
            pass

    def close_destination(self, destination):
        pass

    def _control_word(self, matchobject, cword, param):
        #log.debug('control word %r at index %Xh' % (matchobject.group(), self.index))
        # TODO: according to RTF specs v1.9.1, "Destination changes are legal only immediately after an opening brace ({)"
        # (not counting the special control symbol \*, of course)
        if cword in DESTINATION_CONTROL_WORDS:
            log.debug('%r is a destination control word: starting a new destination at index %Xh' % (cword, self.index))
            self._open_destination(matchobject, cword)
        # call the corresponding user method for additional processing:
        self.control_word(matchobject, cword, param)

    def control_word(self, matchobject, cword, param):
        pass

    def control_symbol(self, matchobject):
        #log.debug('control symbol %r at index %Xh' % (matchobject.group(), self.index))
        pass

    def _text(self, matchobject):
        text = matchobject.group()
        self.current_destination.data += text
        self.text(matchobject, text)

    def text(self, matchobject, text):
        #log.debug('text %r at index %Xh' % (matchobject.group(), self.index))
        pass

    def _bin(self, matchobject, param):
        if param is None:
            log.info('Detected anti-analysis trick: \\bin object without length at index %X' % self.index)
            binlen = 0
        else:
            binlen = int(param)
        # handle negative length
        if binlen < 0:
            log.info('Detected anti-analysis trick: \\bin object with negative length at index %X' % self.index)
            # binlen = int(param.strip('-'))
            # According to my tests, if the bin length is negative,
            # it should be treated as a null length:
            binlen=0
        # ignore optional space after \bin
        if ord(self.data[self.index:self.index + 1]) == ord(' '):
            log.debug('\\bin: ignoring whitespace before data')
            self.index += 1
        log.debug('\\bin: reading %d bytes of binary data' % binlen)
        # TODO: handle length greater than data
        bindata = self.data[self.index:self.index + binlen]
        self.index += binlen
        self.bin(bindata)

    def bin(self, bindata):
        pass

    def _end_of_file(self):
        # log.debug('%Xh Reached End of File')
        # close any group/destination that is still open:
        while self.group_level > 0:
            log.debug('Group Level = %d, closing group' % self.group_level)
            self._close_group()
        self.end_of_file()

    def end_of_file(self):
        pass


class RtfObject(object):
    """
    An object or a file (OLE Package) embedded into an RTF document
    """
    def __init__(self):
        """
        RtfObject constructor
        """
        # start and end index in the RTF file:
        self.start = None
        self.end = None
        # raw object data encoded in hexadecimal, as found in the RTF file:
        self.hexdata = None
        # raw object data in binary form, decoded from hexadecimal
        self.rawdata = None
        # OLE object data (extracted from rawdata)
        self.is_ole = False
        self.oledata = None
        self.format_id = None
        self.class_name = None
        self.oledata_size = None
        # OLE Package data (extracted from oledata)
        self.is_package = False
        self.olepkgdata = None
        self.filename = None
        self.src_path = None
        self.temp_path = None
        # Additional OLE object data
        self.clsid = None
        self.clsid_desc = None




class RtfObjParser(RtfParser):
    """
    Specialized RTF parser to extract OLE objects
    """

    def __init__(self, data):
        super(RtfObjParser, self).__init__(data)
        # list of RtfObjects found
        self.objects = []

    def open_destination(self, destination):
        # TODO: detect when the destination is within an objdata, report as obfuscation
        if destination.cword == b'objdata':
            log.debug('*** Start object data at index %Xh' % destination.start)

    def close_destination(self, destination):
        if destination.cword == b'objdata':
            log.debug('*** Close object data at index %Xh' % self.index)
            rtfobj = RtfObject()
            self.objects.append(rtfobj)
            rtfobj.start = destination.start
            rtfobj.end = destination.end
            # Filter out all whitespaces first (just ignored):
            hexdata1 = destination.data.translate(None, b' \t\r\n\f\v')
            # Then filter out any other non-hex character:
            hexdata = re.sub(b'[^a-fA-F0-9]', b'', hexdata1)
            if len(hexdata) < len(hexdata1):
                # this is only for debugging:
                nonhex = re.sub(b'[a-fA-F0-9]', b'', hexdata1)
                log.debug('Found non-hex chars in hexdata: %r' % nonhex)
            # MS Word accepts an extra hex digit, so we need to trim it if present:
            if len(hexdata) & 1:
                log.debug('Odd length, trimmed last byte.')
                hexdata = hexdata[:-1]
            rtfobj.hexdata = hexdata
            object_data = binascii.unhexlify(hexdata)
            rtfobj.rawdata = object_data
            rtfobj.rawdata_md5 = hashlib.md5(object_data).hexdigest()                    
            # TODO: check if all hex data is extracted properly

            obj = oleobj.OleObject()
            try:
                obj.parse(object_data)
                rtfobj.format_id = obj.format_id
                rtfobj.class_name = obj.class_name
                rtfobj.oledata_size = obj.data_size
                rtfobj.oledata = obj.data
                rtfobj.oledata_md5 = hashlib.md5(obj.data).hexdigest()         
                rtfobj.is_ole = True
                if obj.class_name.lower() == b'package':
                    opkg = oleobj.OleNativeStream(bindata=obj.data,
                                                  package=True)
                    rtfobj.filename = opkg.filename
                    rtfobj.src_path = opkg.src_path
                    rtfobj.temp_path = opkg.temp_path
                    rtfobj.olepkgdata = opkg.data
                    rtfobj.olepkgdata_md5 = hashlib.md5(opkg.data).hexdigest()     
                    rtfobj.is_package = True
                else:
                    if olefile.isOleFile(obj.data):
                        ole = olefile.OleFileIO(obj.data)
                        rtfobj.clsid = ole.root.clsid
                        rtfobj.clsid_desc = clsid.KNOWN_CLSIDS.get(rtfobj.clsid,
                            'unknown CLSID (please report at https://github.com/decalage2/oletools/issues)')
            except:
                pass
                log.debug('*** Not an OLE 1.0 Object')

    def bin(self, bindata):
        if self.current_destination.cword == b'objdata':
            # TODO: keep track of this, because it is unusual and indicates potential obfuscation
            # trick: hexlify binary data, add it to hex data
            self.current_destination.data += binascii.hexlify(bindata)

    def control_word(self, matchobject, cword, param):
        # TODO: extract useful cwords such as objclass
        # TODO: keep track of cwords inside objdata, because it is unusual and indicates potential obfuscation
        # TODO: same with control symbols, and opening bracket
        # log.debug('- Control word "%s", param=%s, level=%d' % (cword, param, self.group_level))
        pass

    def control_symbol(self, matchobject):
        # log.debug('control symbol %r at index %Xh' % (matchobject.group(), self.index))
        symbol = matchobject.group()[1:2]
        if symbol == b"'":
            # read the two hex digits following "\'" - which can be any characters, not just hex digits
            # (because within an objdata destination, they are simply ignored)
            hexdigits = self.data[self.index+2:self.index+4]
            # print(hexdigits)
            # move the index two bytes forward
            self.index += 2
            if self.current_destination.cword == b'objdata':
                # Here's the tricky part: there is a bug in the MS Word RTF parser at least
                # until Word 2016, that removes the last hex digit before the \'hh control
                # symbol, ONLY IF the number of hex digits read so far is odd.
                # So to emulate that bug, we have to clean the data read so far by keeping
                # only the hex digits:
                # Filter out any non-hex character:
                self.current_destination.data = re.sub(b'[^a-fA-F0-9]', b'', self.current_destination.data)
                if len(self.current_destination.data) & 1 == 1:
                    # If the number of hex digits is odd, remove the last one:
                    self.current_destination.data = self.current_destination.data[:-1]


#=== FUNCTIONS ===============================================================
def process(data):
    rtfp = RtfObjParser(data)
    rtfp.parse()
    results = list()
    for rtfobj in rtfp.objects:
        r = dict()
        if rtfobj.is_ole:
            r['is_ole'] = True
            r['format_id'] = rtfobj.format_id
            if rtfobj.format_id == oleobj.OleObject.TYPE_EMBEDDED:
                r['format_type'] = 'Embedded'
            elif rtfobj.format_id == oleobj.OleObject.TYPE_LINKED:
                r['format_type'] = 'Linked'
            else:
                r['format_type'] = 'Unknown'
            r['class_name'] = rtfobj.class_name
            if r['class_name']:
                r['class_name'] = r['class_name'].decode('utf8', 'ignore')
            # if the object is linked and not embedded, data_size=None:
            if rtfobj.oledata_size is None:
                r['data_size'] = -1
            else:
                r['data_size'] = rtfobj.oledata_size
            if rtfobj.is_package:
                r['is_package'] = True
                r['file_name'] = rtfobj.filename
                r['src_path'] = rtfobj.src_path
                r['temp_path'] = rtfobj.temp_path
                r['md5'] = rtfobj.olepkgdata_md5
            else:
                r['md5'] = rtfobj.oledata_md5
            if rtfobj.clsid is not None:
                r['clsid'] = rtfobj.clsid
                r['clsid_desc'] = rtfobj.clsid_desc
            # Detect OLE2Link exploit
            # http://www.kb.cert.org/vuls/id/921560
            if rtfobj.class_name == b'OLE2Link':
                found_list =  re.findall(r'[a-fA-F0-9\x0D\x0A]{128,}',data)
                urls = []
                for item in found_list:
                    try:
                        temp = item.replace("\x0D\x0A","").decode("hex")
                    except:
                        continue
                    pat = re.compile(r'(?:[\x20-\x7E][\x00]){3,}')
                    words = [w.decode('utf-16le') for w in pat.findall(temp)]
                    for w in words:
                        if "http" in w:
                            urls.append(w)
                urls = sorted(set(urls))
                if urls:
                    r['urls'] = urls
            results.append(r)

    return results



# -*- coding: utf-8 -*-
# @Time    : 2020/12/23 01:59:09 
# @Author  : ddvv
# @Site    : https://ddvvmmzz.github.io
# @File    : __init__.py 
# @Software: Visual Studio Code


from mmpi.core.plugins import enumerate_plugins
from mmpi.common.abstracts import Processing

plugins = enumerate_plugins(
    __file__, "mmpi.processing", globals(), Processing
)



# -*- coding: utf-8 -*-
# @Time    : 2020/12/23 01:59:33 
# @Author  : ddvv
# @Site    : https://ddvvmmzz.github.io
# @File    : __init__.py 
# @Software: Visual Studio Code


from mmpi import processing

signatures = []

# Don't include machinery here as its data structure is different from the
# other plugins - of which multiple are in use at any time.
plugins = {
    "processing": processing.plugins,
    "signatures": signatures,
}

from mmpi.main import mmpi


# -*- coding: utf-8 -*-
# @Time    : 2020/12/31 16:05:02 
# @Author  : ddvv
# @Site    : https://ddvvmmzz.github.io
# @File    : __init__.py 
# @Software: Visual Studio Code


from ..compat import enumerate_signatures

plugins = enumerate_signatures(
    __file__, "rtf", globals(), dict()
)


# -*- coding: utf-8 -*-
# @Time    : 2020/12/21 21:13:06
# @Author  : ddvv
# @Site    : https://ddvvmmzz.github.io
# @File    : oleparser.py
# @Software: Visual Studio Code


import logging

from mmpi.common.abstracts import Processing
from mmpi.common.exceptions import ProcessingError
from mmpi.common.olevba import VBA_RUN
# from mmpi.common.filetypes import get_magic_type


log = logging.getLogger(__name__)


class OleParser(Processing):

    @classmethod
    def init_once(cls):
        log.debug("Initializing OleParser...")

        try:
            # nothing to do
            cls.vr = VBA_RUN()
        except Exception as e:
            raise ProcessingError(
                "There was a syntax error in one or more OleParser: %s" % e
            )

    def run(self):
        self.key = "ole"
        if self.key == self.file_type:
            vba = {'vba': list()}
            info = {'infos': list(), 'datas': list(), 'report': vba}

            if self.file_content:
                # magic_type = get_magic_type(self.file_content)
                try:
                    if self.vr.is_ole(self.file_content):
                        for data in self.vr.parse(self.file_content):
                            vba_code = data.get('vba_code')
                            info['datas'].append({'file_type': 'vba', 'filename': data.get(
                                'vba_filename'), 'filesize': len(vba_code), 'content': vba_code})
                            vba['vba'].append(
                                {'size': len(vba_code), 'code': vba_code.decode()})
                    else:
                        # maybe rtf
                        info['datas'].append(
                            {'file_type': 'rtf', 'content': self.file_content})
                    if info.get('infos', []) or info.get('datas', []):
                        return info
                except Exception as e:
                    # magic_type = get_magic_type(self.file_content)
                    magic_type = ''
                    info['infos'].append(
                        {'type': self.file_type, 'magic_type': magic_type})
                    return info
        return None



# -*- coding: utf-8 -*-
# @Time    : 2020/12/30 00:21:10
# @Author  : ddvv
# @Site    : https://ddvvmmzz.github.io
# @File    : rtfparser.py
# @Software: Visual Studio Code


import logging

from mmpi.common.abstracts import Processing
from mmpi.common.exceptions import ProcessingError
from mmpi.common.rtfextract import process


log = logging.getLogger(__name__)


class RTFParser(Processing):

    @classmethod
    def init_once(cls):
        log.debug("Initializing RTFParser...")

        try:
            # nothing to do
            pass
        except Exception as e:
            raise ProcessingError(
                "There was a syntax error in one or more RTFParser: %s" % e
            )

    def run(self):
        self.key = "rtf"
        if self.key == self.file_type:
            rtf = {'rtf': list()}
            info = {'infos': list(), 'datas': list(), 'report': rtf}

            if self.file_content:
                # magic_type = get_magic_type(self.file_content)
                try:
                    datas = process(self.file_content)
                    rtf['rtf'] = datas
                    info['infos'] = datas
                    if info.get('infos', []) or info.get('datas', []) or rtf.get('rtf', {}):
                        return info
                except Exception as e:
                    pass
        return None
